import os
import itertools

import pandas as pd
import pyranges as pr

import handygenome
import handygenome.refgenome.refgenome as refgenome


# curated blacklist

CURATED_BLACKLIST = refgenome.RefverDict({
    'GRCh37': {
        'peri_centromere': [
            ('10', 38_000_000, 42_523_247),
            ('10', 42_527_152, 42_546_687),
            # 42_546_687 - 42_596_687 : N
            ('10', 42_596_687, 42_818_502),

            ('7', 61_967_157, 62_010_018),
            ('7', 62_015_012, 62_054_190),

            ('11', 51_583_335, 51_594_205),

            ('Y', 13_400_000, 13_573_805),
            ('Y', 13_634_339, 13_748_578),

            ('8', 43_092_656, 43_100_000),

            ('2', 89_830_436, 89_880_102),
        ],
        'non_centromere': [
            ('1', 144_810_724, 145_382_727),
            ('2', 33_091_515, 33_092_197),
            ('2', 33_093_197, 33_095_590),
            ('2', 33_141_000, 33_141_692),
        ],
    },
})


def make_blacklist_gr(data, refver):
    df_data = list(itertools.chain.from_iterable(data.values()))
    df = pd.DataFrame.from_records(df_data, columns=['Chromosome', 'Start', 'End'])
    assert set(df['Chromosome']).issubset(refgenome.get_chromdict(refver).contigs)
    gr = pr.PyRanges(df).merge().sort()

    return gr


CURATED_BLACKLIST_GRS = refgenome.RefverDict(
    {refver: make_blacklist_gr(data, refver) for refver, data in CURATED_BLACKLIST.items()}
)


# blacklist generated with excessive depth region calculation with normal bams

HIGHDEPTH_BLACKLIST_PATH = os.path.join(handygenome.DIRS['data'], 'custom_blacklist.tsv.gz')
HIGHDEPTH_BLACKLIST_DF = pd.read_csv(
    HIGHDEPTH_BLACKLIST_PATH, 
    sep='\t',
    header=0,
    usecols=['Chromosome', 'Start', 'End'],
    dtype={'Chromosome': str, 'Start': int, 'End': int},
)
HIGHDEPTH_BLACKLIST_GR = pr.PyRanges(HIGHDEPTH_BLACKLIST_DF)






import os
import pickle

import handygenome.logutils as logutils
import handygenome.tools as tools
import handygenome.refgenome.refgenome as refgenome
import handygenome.fastahandler as fastahandler
import handygenome.publicdb.ncbi as libncbi
import handygenome.publicdb.ncbi_cache as libncbicache
from handygenome.genomedf.genomedf import GenomeDataFrame as GDF


##############
# exceptions #
##############

class PARUnavailableError(Exception):
    pass


############
# N region #
############

def get_Nregion_path(refver):
    return os.path.join(refgenome.get_datadir(refver), 'Nregion.bed.gz')


def write_Nregion_file(refver, path):
    logutils.log(f'Writing a new N region file for refver {refver}')

    fasta = refgenome.get_fasta(refver)
    chroms, start0s, end0s = fastahandler.make_N_region_coords(fasta)
    Nregion_gdf = GDF.from_data(refver=refver, chroms=chroms, start0s=start0s, end0s=end0s)
    Nregion_gdf.write_tsv(path)
    logutils.log(f'Finished writing a new N region file for refver {refver}')


def get_Nregion_gdf(refver):
    Nregion_path = get_Nregion_path(refver)
    if not os.path.exists(Nregion_path):
        write_Nregion_file(refver, Nregion_path)
    return GDF.read_tsv(Nregion_path, refver)


#######
# PAR #
#######

def get_assembly_regions_df(refver, force_download=False):
    RefSeq_refver, species = refgenome.get_RefSeq_standard(refver)
    return libncbi.parse_assembly_regions(
        libncbicache.get_assembly_regions_path(
            RefSeq_refver, species, force_download=force_download,
        )
    )


def get_par_gdf(refver, force_download=False):
    df = get_assembly_regions_df(refver, force_download=force_download)
    par_subdf = df.loc[df['Scaffold-Role'] == 'PAR', :]
    if par_subdf.shape[0] == 0:
        raise PARUnavailableError(f'PAR is not available for this reference version.')

    chrom_converter = refgenome.get_chrom_converter(refver)
    chroms = [chrom_converter[x] for x in par_subdf['Chromosome']]

    # assume the coordinates are 1-based both closed
    start0s = par_subdf['Chromosome-Start'] - 1
    end0s = par_subdf['Chromosome-Stop']

    return GDF.from_data(
        refver=refver,
        chroms=chroms,
        start0s=start0s,
        end0s=end0s,
    )


###############
# GC fraction #
###############

def get_gcfraction_path(refver, binsize):
    refver_datadir = refgenome.get_datadir(refver)
    gcfile_path = os.path.join(refver_datadir, f'gcfraction_bin{binsize}.pickle')
    return gcfile_path


def write_gcfraction(refver, binsize):
    logutils.log(f'Writing a new GC fraction file for refver {refver} and binsize {binsize}')

    gcfile_path = get_gcfraction_path(refver, binsize)
    fasta = refgenome.get_fasta(refver)
    all_regions = GDF.all_regions(refver, assembled_only=False).window(binsize)
    all_regions.sort()

    chroms = all_regions['Chromosome']
    start0s = all_regions['Start']
    end0s = all_regions['End']
    gcs = fastahandler.calculate_gc_fractions(chroms, start0s, end0s, fasta, window=None, as_array=True)

    gc_gdf = GDF.from_data(
        refver=refver, chroms=chroms, start0s=start0s, end0s=end0s, GC=gcs,
    )

    with open(gcfile_path, 'wb') as outfile:
        pickle.dump(gc_gdf, outfile)
    
    
def load_gcfraction(refver, binsize):
    gcfile_path = get_gcfraction_path(refver, binsize)
    if not os.path.exists(gcfile_path):
        write_gcfraction(refver, binsize)

    with open(gcfile_path, 'rb') as infile:
        return pickle.load(infile)
    




import os
import itertools
import warnings
import textwrap

import pysam

import handygenome.refgenome.refgenome as refgenome
import handygenome.workflow as workflow
import handygenome.variant.varianthandler as varianthandler
import handygenome.vcfeditor.headerhandler as headerhandler
import handygenome.vcfeditor.initvcf as initvcf
import handygenome.vcfeditor.misc as vcfmisc
import handygenome.variant.vcfspec as libvcfspec


LOGGER = workflow.get_logger()


def sanity_check_args(infile_path_list, outfile_path, isec, isec_indices,
                      outfile_must_not_exist):
    infile_path_list = workflow.arghandler_infilelist(infile_path_list)

    arghandler = workflow.get_arghandler_outfile(outfile_must_not_exist)
    outfile_path = arghandler(outfile_path)

    sanity_check_isec_indices(isec_indices, len(infile_path_list))

    return infile_path_list, outfile_path


def sanity_check_isec_indices(isec_indices, n_vcf):
    e_msg = textwrap.dedent(f"""\
        'isec_indices' must be a list composed of 0 or 1, \
        with the length the same as the number of vcf inputs.""")
    if isec_indices is not None:
        if len(isec_indices) != n_vcf:
            raise Exception(e_msg)
        elif not set(isec_indices).issubset({0, 1}):
            raise Exception(e_msg)


############################################


def isec_indices_to_bool(isec_indices, n_vcf):
    if isec_indices is None:
        isec_indices_bool = list(itertools.repeat(True, n_vcf))
    else:
        isec_indices_bool = [bool(x) for x in isec_indices]

    return isec_indices_bool


def load_vcf_data(vcf_list, remove_infoformat):
    """Returns:
        [
            {vcfspec, vcfspec, ... },
            {vcfspec, vcfspec, ... },
            ...,
        ] (same length as len(infile_path_list)) 
    """

    raw_vcfspecs = list()
    if remove_infoformat:
        vr_dict = None
    else:
        vr_dict = dict()
    for in_vcf in vcf_list:
        subset = set()
        for vr in in_vcf.fetch():
            vcfspec = libvcfspec.Vcfspec.from_vr(vr)
            #vcfspec = varianthandler.get_vcfspec(vr)
            subset.add(vcfspec)
            if not remove_infoformat:
                vr_dict.setdefault(vcfspec, list())
                vr_dict[vcfspec].append(vr)

        raw_vcfspecs.append(subset)

    return raw_vcfspecs, vr_dict


#def load_vcf_data_vcfpath(infile_path_list):
#    vcf_list = [pysam.VariantFile(x) for x in infile_path_list]
#    raw_vcfspecs, vr_dict = load_vcf_data(vcf_list)
#    for vcf in vcf_list:
#        vcf.close()
#
#    return raw_vcfspecs, vr_dict


########################################


def get_merged_header(infile_path_list):
    """Any conflicting INFO or FORMAT keys 
    (with regard to any of Type, Number, or Description) are discarded."""

    header_list = list()
    for vcf_path in infile_path_list:
        with pysam.VariantFile(vcf_path, 'r') as in_vcf:
            header_list.append(in_vcf.header)
    #compatible = headerhandler.check_header_compatibility(header_list)
    merged_header = headerhandler.merge_vcfheaders(header_list)

    return merged_header


def get_output_vcfspecs_isec(raw_vcfspecs, isec_indices_bool, chromdict):
    vcfspecs_included = set.intersection( 
        *itertools.compress(raw_vcfspecs, isec_indices_bool)
    )
    for subset in itertools.compress(
        raw_vcfspecs, [not x for x in isec_indices_bool]
    ):
        vcfspecs_included.difference_update(subset)
    output_vcfspecs = sorted(
        vcfspecs_included, 
        #key=common.get_vcfspec_sortkey(chromdict)
        key=(lambda x: x.get_sortkey(chromdict)),
    )
    return output_vcfspecs


def get_output_vcfspecs_union(raw_vcfspecs, chromdict):
    vcfspec_set = set.union(*raw_vcfspecs)
    output_vcfspecs = sorted(
        vcfspec_set, 
        #key=common.get_vcfspec_sortkey(chromdict),
        key=(lambda x: x.get_sortkey(chromdict)),
    )

    return output_vcfspecs


def merge_vr_dict(vr_dict, output_header, output_vcfspecs, logger):
    output_vr_list = list()
    NR = 0
    for vcfspec in output_vcfspecs:
        NR += 1
        merged_vr = varianthandler.merge(vr_dict[vcfspec], output_header)
        output_vr_list.append(merged_vr)
        del vr_dict[vcfspec]
        if NR % 50_000 == 0:
            logger.info(f'{NR:,} merged variant records created')

    return output_vr_list


def get_output_vrs_wo_infoformat(output_vcfspecs, output_header):
    output_vr_list = list()
    for vcfspec in output_vcfspecs:
        new_vr = output_header.new_record()
        varianthandler.apply_vcfspec(new_vr, vcfspec)
        output_vr_list.append(new_vr)

    return output_vr_list


def write(output_vr_list, output_vcfspecs, outfile_path, mode_pysam, 
          output_header):
    with pysam.VariantFile(outfile_path, mode=mode_pysam, 
                           header=output_header) as out_vcf:
        for vr in output_vr_list:
            out_vcf.write(vr)


#########################################


def main_file(
    infile_path_list, 
    outfile_path, 
    fasta_path, 
    remove_infoformat=False,
    isec=False, 
    isec_indices=None, 
    mode_bcftools='z', 
    mode_pysam=None,
    outfile_must_not_exist='ask',
    logger=None,
):
    if logger is None:
        logger = LOGGER

    logger.info('Beginning')

    # set basic parameters
    infile_path_list, outfile_path = sanity_check_args(
        infile_path_list, 
        outfile_path, 
        isec, 
        isec_indices, 
        outfile_must_not_exist,
    )
    isec_indices_bool = isec_indices_to_bool(isec_indices, len(infile_path_list))
    mode_pysam = vcfmisc.write_mode_arghandler(mode_bcftools, mode_pysam)
    fasta = pysam.FastaFile(fasta_path)
    chromdict = refgenome.ChromDict.from_fasta(fasta)
    if remove_infoformat:
        output_header = initvcf.create_header(chromdict=chromdict)
    else:
        output_header = get_merged_header(infile_path_list)

    # load input files data
    vcf_list = [pysam.VariantFile(x) for x in infile_path_list]

    logger.info('Loading input vcf data')
    raw_vcfspecs, vr_dict = load_vcf_data(vcf_list, remove_infoformat)

    # create output_vcfspecs
    logger.info('Getting union/intersection of vcfspecs')
    if isec:
        output_vcfspecs = get_output_vcfspecs_isec(raw_vcfspecs, isec_indices_bool, chromdict)
    else:
        output_vcfspecs = get_output_vcfspecs_union(raw_vcfspecs, chromdict)
    logger.info(f'{len(output_vcfspecs):,} variant records are to be written')

    # create output_vr_list
    if remove_infoformat:
        output_vr_list = get_output_vrs_wo_infoformat(output_vcfspecs, output_header)
    else:
        logger.info('Merging overlapping variant records. This step may take a while.')
        output_vr_list = merge_vr_dict(vr_dict, output_header, output_vcfspecs, logger)

    # write final result
    logger.info('Writing final result')
    write(output_vr_list, output_vcfspecs, outfile_path, mode_pysam, output_header)

    for vcf in vcf_list:
        vcf.close()
    fasta.close()


#def main_vcfplus(vcfplus_list, isec=False, isec_indices=None, merge=True):
#    #assert all(isinstance(x, vcfplus.VcfPlus) for x in vcfplus_list)
#    def get_vp_list(output_vcfspecs, vr_dict, fasta, chromdict):
#        vp_list = list()
#        for vcfspec in output_vcfspecs:
#            for vr in vr_dict[vcfspec]:
#                vp = variantplus.VariantPlus(vr, fasta, chromdict)
#                vp_list.append(vp)
#
#        return vp_list
#
#    sanity_check_isec_indices(isec_indices, len(vcfplus_list))
#    chromdict = vcfplus_list[0].chromdict
#    fasta = vcfplus_list[0].fasta
#    isec_indices_bool = isec_indices_to_bool(isec_indices, len(vcfplus_list))
#
#    merged_header = merge_pysamhdr(vcfp.vcf.header for vcfp in vcfplus_list)
#    raw_vcfspecs = load_vcf_data(vcf_list)
#    output_vcfspecs = vcfmergelib.get_output_vcfspecs(raw_vcfspecs, isec, isec_indices_bool, chromdict)
#    vr_dict = vcfmergelib.get_vr_dict(vcfplus_list, output_vcfspecs, merged_header, isec_indices_bool, merge)
#    vp_list = get_vp_list(output_vcfspecs, vr_dict, fasta, chromdict)
#
#    vcfp = vcfplus.VcfPlus(fasta = fasta)
#    vcfp.set_header(header = merged_header)
#    vcfp.init_vp(vp_list = vp_list)
#
#    return vcfp



import itertools
import collections
import functools

import pyranges as pr

import importlib

top_package_name = __name__.split(".")[0]
common = importlib.import_module(".".join([top_package_name, "common"]))
readhandler = importlib.import_module(
    ".".join([top_package_name, "read", "readhandler"])
)


MAX_READ_LENGTH = 151
DEFAULT_FETCHEDREADS_ADDITION_PADDING = 200
DEFAULT_FETCHEDREFSEQ_ADDITION_PADDING = 200


def cluster_fetch_arghandler(query_start0, query_end0, db_start0, db_end0):
    if (
        ((query_end0 is not None) and (query_end0 <= db_start0)) or
        ((query_start0 is not None) and (query_start0 >= db_end0))
    ):
        raise Exception(f'Input fetch range does not overlap with database range.')

    if query_start0 is None:
        result_start0 = db_start0
    else:
        result_start0 = max(query_start0, db_start0)

    if query_end0 is None:
        result_end0 = db_end0
    else:
        result_end0 = min(query_end0, db_end0)

    return result_start0, result_end0


class ClusterBase:
    def get_id(self):
        return (self.chrom, self.start0, self.end0)


class SinglechromBase:
    @property
    def gr(self):
        starts = list()
        ends = list()
        ids = list()
        for x in self.clusters:
            starts.append(x.start0)
            ends.append(x.end0)
            ids.append(x.get_id())
        chroms = [self.chrom] * len(self.clusters)
        return pr.from_dict(
            {'Chromosome': chroms, 'Start': starts, 'End': ends, 'Id': ids}
        )

    def get_clusters_by_ids(self, ids):
        result = dict()
        for x in self.clusters:
            id = x.get_id()
            if id in ids:
                result[id] = x
        return result

    def sort_clusters(self):
        self.clusters.sort(key=(lambda x: x.fetch_start0))

    def merge(self):
        new_clusters = list()
        self.sort_clusters()

        def unit_func(cluster1, cluster2):
            if cluster1.end0 >= cluster2.start0:
                return cluster1.merge(cluster2)
            else:
                new_clusters.append(cluster1)
                return cluster2

        final = functools.reduce(unit_func, self.clusters)
        new_clusters.append(final)
        self.clusters = new_clusters

    def add(self, start0, end0):
        if len(self.clusters) == 0:
            self.clusters.append(self.spawn_cluster_from_fetch(start0, end0))
        else:
            self_gr = self.gr
            input_gr = pr.PyRanges(chromosomes=[self.chrom], starts=[start0], ends=[end0])
            subtracted_gr = input_gr.subtract(self_gr)
                # When there is nothing left after subtraction, "subtracted_gr" becomes an empty PyRanges. 
            nearest_gr = subtracted_gr.nearest(self_gr)
                # "nearest_gr" becomes also an empty PyRanges. 
                # Iteration with .df.iterrows() results in an empty iteration.
            # get adjacent clusters
            adj_cluster_ids = list()
            for idx, row in nearest_gr.df.iterrows():
                if abs(row['Distance']) == 1:
                    adj_cluster_ids.append(row['Id'])
            adj_clusters = self.get_clusters_by_ids(adj_cluster_ids)
            # do read addition
            for idx, row in nearest_gr.df.iterrows():
                if abs(row['Distance']) == 1:
                    cluster = adj_clusters[row['Id']]
                    added_region_width = row['End'] - row['Start']
                    if row['Start'] == row['End_b']:
                        cluster.extend_rightward(row['End'] - row['Start'])
                    elif row['End'] == row['Start_b']:
                        cluster.extend_leftward(row['End'] - row['Start'])
                else:
                    self.clusters.append(self.spawn_cluster_from_fetch(row['Start'], row['End']))

            if len(adj_clusters) > 0:
                self.merge()

    def get_valid_cluster(self, start0, end0):
        """When query range is not contained within database, 'add' is automatically done."""
        input_gr = pr.PyRanges(chromosomes=[self.chrom], starts=[start0], ends=[end0])
        joined_gr = input_gr.join(self.gr, how=None, apply_strand_suffix=False)

        valid_cluster_ids = list()
        for idx, row in joined_gr.df.iterrows():
            if row['Start_b'] <= row['Start'] and row['End_b'] >= row['End']:
                valid_cluster_ids.append(row['Id'])

        if len(valid_cluster_ids) == 0:
            self.add(start0 - self.addition_padding, end0 + self.addition_padding)
            return self.get_valid_cluster(start0, end0)
        elif len(valid_cluster_ids) == 1:
            cluster_id = valid_cluster_ids[0]
            cluster = self.get_clusters_by_ids([cluster_id])[cluster_id]
            return cluster
        else:
            raise Exception(f'There are more than one clusters containing input fetch range.')


##############################################


class FetchedReadsCluster(ClusterBase):
    def __init__(self, bam, chrom, readfilter=None):
        self.bam = bam
        self.chrom = chrom

        if readfilter is None:
            self.readfilter = readhandler.readfilter_pileup
        else:
            self.readfilter = readfilter

        self.fetch_start0 = None
        self.fetch_end0 = None
        self.dict = collections.OrderedDict()

    def __repr__(self):
        return f'<FetchedReadsCluster object (chrom={self.chrom}, fetch_start0={self.fetch_start0:,}, fetch_end0={self.fetch_end0:,})>'

    @classmethod
    def from_fetch(cls, bam, chrom, start0, end0, readfilter=None):
        result = cls(bam, chrom, readfilter)
        result.fetch_start0 = start0
        result.fetch_end0 = end0

        for read in readhandler.get_fetch(
            result.bam, result.chrom, start0, end0, readfilter=result.readfilter
        ):
            read_uid = readhandler.get_uid(read)
            result.dict[read_uid] = read

        return result

    ###########
    # yielder #
    ###########
    @staticmethod
    def yielder_with_uid(uid, read):
        return uid, read

    @staticmethod
    def yielder_without_uid(uid, read):
        return read

    def get_yielder(self, with_uid):
        if with_uid:
            return self.yielder_with_uid
        else:
            return self.yielder_without_uid
    ###########

    @property
    def start0(self):
        return self.fetch_start0

    @property
    def end0(self):
        return self.fetch_end0

    @property
    def fetch_range0(self):
        return range(self.fetch_start0, self.fetch_end0)

    @property
    def readspan_start0(self):
        if len(self.dict) == 0:
            return None
        else:
            return min(read.reference_start for read in self.dict.values())

    @property
    def readspan_end0(self):
        if len(self.dict) == 0:
            return None
        else:
            return max(read.reference_end for read in self.dict.values())

#    def _check_uid_uniqueness(self):
#        if len(set(self.uids)) != len(self.uids):
#            reads_string = "\n".join(x.to_string() for x in self.list)
#            uids_string = "\n".join(x for x in self.uids)
#            raise Exception(
#                f"Duplicate read uids.\n"
#                f"Read list:\n{reads_string}.\n"
#                f"uid list:\n{uids_string}."
#            )

    def get_id(self):
        return (self.chrom, self.fetch_start0, self.fetch_end0)

    def get_read(self, read_uid):
        return self.dict[read_uid]

    def copy(self):
        result = self.__class__(self.bam, self.chrom)
        result.fetch_start0 = self.fetch_start0
        result.fetch_end0 = self.fetch_end0
        result.dict.update(self.dict)
        return result

    # merge, extend #
    def merge(self, other):
        if (other.fetch_start0 > self.fetch_end0) or (self.fetch_start0 > other.fetch_end0):
            raise Exception(f'Fetch ranges of two objects do not overlap.')

        if self.fetch_start0 <= other.fetch_start0 and self.fetch_end0 >= other.fetch_end0:
            return self.copy()
        elif other.fetch_start0 <= self.fetch_start0 and other.fetch_end0 >= self.fetch_end0:
            return other.copy()
        else:
            # now self.fetch_start0 != other.fetch_start0
            if self.fetch_start0 < other.fetch_start0:
                left = self
                right = other
            else:
                left = other
                right = self

            result = self.__class__(self.bam, self.chrom)
            result.fetch_start0 = left.fetch_start0
            result.fetch_end0 = right.fetch_end0

            left_only_width = right.fetch_start0 - left.fetch_start0
            right_only_width = right.fetch_end0 - left.fetch_end0
            if left_only_width >= right_only_width:
                result.dict.update(left.dict)

                buffer = list(
                    itertools.takewhile(
                        (lambda x: x[0] not in left.dict),
                        reversed(right.dict.items())
                    )
                )
                result.dict.update(reversed(buffer))
            else:
                result.dict.update(
                    itertools.takewhile(
                        (lambda x: x[0] not in right.dict),
                        left.dict.items()
                    )
                )
                result.dict.update(right.dict)

            return result

    def extend_rightward(self, width):
        for read, uid in itertools.dropwhile(
            (lambda x: x[1] in self.dict),
            readhandler.get_fetch(self.bam, self.chrom, self.fetch_end0, self.fetch_end0 + width, readfilter=self.readfilter, with_uid=True),
        ):
            self.dict[uid] = read

        self.fetch_end0 += width

    def extend_leftward(self, width):
        buffer = list(
            itertools.takewhile(
                (lambda x: x[1] not in self.dict),
                readhandler.get_fetch(self.bam, self.chrom, self.fetch_start0 - width, self.fetch_start0, readfilter=self.readfilter, with_uid=True),
            )
        )
        for read, uid in reversed(buffer):
            self.dict[uid] = read
            self.dict.move_to_end(uid, last=False)

        self.fetch_start0 -= width
            
    # fetch #
    def fetch(self, start0=None, end0=None, with_uid=False):
        if len(self.dict) == 0:
            return iter(())
        else:
            start0, end0 = cluster_fetch_arghandler(start0, end0, self.fetch_start0, self.fetch_end0)
            if start0 >= end0:
                return iter(())
            else:
                yielder = self.get_yielder(with_uid)

                for uid, read in self.dict.items():
                    if read.reference_start >= end0:
                        break
                    if read.reference_end > start0:
                        yield yielder(uid, read)

    def _fetch_forward(self, fetch_start0, fetch_end0, with_uid=False):
        yielder = self.get_yielder(with_uid)
            
        is_area1 = False
        is_area2 = False
        for uid, read in self.dict.items():
            # pre-area1
            if not is_area1:
                # check if within area1
                if read.reference_start > fetch_start0 - MAX_READ_LENGTH:
                    is_area1 = True
                else:
                    # job pre-area1 (null)
                    # continue
                    continue
            # area1
            if not is_area2:
                # check if within area2
                if read.reference_start >= fetch_start0:
                    is_area2 = True
                else:
                    # job area1
                    if read.reference_end > fetch_start0:
                        yield yielder(uid, read)
                    # continue
                    continue
            # check if beyond area2
            if read.reference_start >= fetch_end0:
                break
            # job area2
            yield yielder(uid, read)
                
    def _fetch_reverse(self, fetch_start0, fetch_end0, with_uid=False):
        yielder = self.get_yielder(with_uid)
            
        is_area1 = False
        is_area2 = False
        for uid, read in reversed(self.dict.items()):
        #for uid, read in zip(reversed(self.uids), reversed(self.list)):
            # post-area2
            if not is_area2:
                # check if within area2
                if read.reference_start < fetch_end0:
                    is_area2 = True
                else:
                    # job post-area2 (null)
                    # continue
                    continue
            # area2
            if not is_area1:
                # check if within area1
                if read.reference_start < fetch_start0:
                    is_area1 = True
                else:
                    # job area2
                    yield yielder(uid, read)
                    # continue
                    continue
            # check if within pre-area1
            if read.reference_start <= fetch_start0 - MAX_READ_LENGTH:
                break
            # job area1
            if read.reference_end > fetch_start0:
                yield yielder(uid, read)
                                    
    def fetch_choosing_side(self, start0=None, end0=None, with_uid=False):
        if len(self.dict) == 0:
            return iter(())
        else:
            start0, end0 = cluster_fetch_arghandler(start0, end0, self.fetch_start0, self.fetch_end0)
            if start0 >= end0:
                return iter(())
            else:
                # return self._fetch_forward(start0, end0, with_uid=with_uid)                
                dist_from_start = start0 - self.fetch_start0
                dist_from_end = self.fetch_end0 - end0

                if dist_from_start <= dist_from_end:
                    return self._fetch_forward(
                        start0, end0, with_uid=with_uid
                    )
                else:
                    return reversed(
                        tuple(
                            self._fetch_reverse(
                                start0, end0, with_uid=with_uid
                            )
                        )
                    )


class FetchedReadsSinglechrom(SinglechromBase):
    def __init__(self, chrom, bam, readfilter=None, addition_padding=DEFAULT_FETCHEDREADS_ADDITION_PADDING):
        self.chrom = chrom
        self.bam = bam
        self.addition_padding = addition_padding

        if readfilter is None:
            self.readfilter = readhandler.readfilter_pileup
        else:
            self.readfilter = readfilter

        self.clusters = list()

    def __repr__(self):
        return f'<FetchedReadsSinglechrom object (chrom={self.chrom}, clusters={self.clusters})>'

    @classmethod
    def from_fetch(cls, bam, chrom, start0, end0, readfilter=None, addition_padding=DEFAULT_FETCHEDREADS_ADDITION_PADDING):
        result = cls(chrom, bam, readfilter, addition_padding)
        result.clusters.append(FetchedReadsCluster.from_fetch(result.bam, result.chrom, start0, end0, result.readfilter))
        return result

#        return pr.PyRanges(
#            chromosomes=([self.chrom] * len(self.clusters)),
#            starts=[x.fetch_start0 for x in self.clusters],
#            ends=[x.fetch_end0 for x in self.clusters],
#        )

#    def get_clusters_by_ids(self, ids):
#        result = dict()
#        for x in self.clusters:
#            id = x.get_id()
#            if id in ids:
#                result[id] = x
#        return result

    def spawn_cluster_from_fetch(self, start0, end0):
        return FetchedReadsCluster.from_fetch(self.bam, self.chrom, start0, end0, self.readfilter)

    def get_read(self, read_uid):
        found_read = False
        for frcluster in self.clusters:
            try:
                read = frcluster.dict[read_uid]
            except KeyError:
                continue
            else:
                found_read = True
                break

        if not found_read:
            raise Exception(f'Given ReadUID is not present.')

        return read

    def iter_reads(self):
        return itertools.chain.from_iterable(x.dict.items() for x in self.clusters)

#    def sort_clusters(self):
#        self.clusters.sort(key=(lambda x: x.fetch_start0))

    def add_reads(self, start0, end0):
        self.add(start0, end0)

    def fetch(self, start0, end0, with_uid=False):
        valid_cluster = self.get_valid_cluster(start0, end0)
        return valid_cluster.fetch(start0, end0, with_uid=with_uid)

    def naive_fetch(self, start0, end0, with_uid=False):
        input_gr = pr.PyRanges(chromosomes=[self.chrom], starts=[start0], ends=[end0])
        joined_gr = input_gr.join(self.gr, how=None, apply_strand_suffix=False)

        valid_frcluster_ids = list()
        for idx, row in joined_gr.df.iterrows():
            if row['Start_b'] <= row['Start'] and row['End_b'] >= row['End']:
                valid_frcluster_ids.append(row['Id'])

        if len(valid_frcluster_ids) == 0:
            raise Exception(f'Input fetch range is not fully contained within database.')
        if len(valid_frcluster_ids) > 1:
            raise Exception(f'There are more than one FetchedReadsCluster objects containing input fetch range.')

        frcluster_id = valid_frcluster_ids[0]
        frcluster = self.get_clusters_by_ids([frcluster_id])[frcluster_id]
        return frcluster.fetch(start0, end0, with_uid=with_uid)

    def check_fully_includes(self, start0, end0):
        input_gr = pr.PyRanges(chromosomes=[self.chrom], starts=[start0], ends=[end0])
        intersect_gr = input_gr.intersect(self.gr, how='containment')
        return not intersect_gr.empty

    def choose_cluster(self, start0, end0):
        self_gr = self.gr
        input_gr = pr.PyRanges(chromosomes=[self.chrom], starts=[start0], ends=[end0])
        joined_gr = input_gr.join(self_gr, how=None, apply_strand_suffix=False)
        if joined_gr.df.shape[0] == 0:
            return None
        else:
            overlapping_cluster_ids = {row['Id'] for idx, row in joined_gr.df.iterrows()}
            overlapping_clusters = list()
            for idx, row in self_gr.df.iterrows():
                if row['Id'] in overlapping_cluster_ids:
                    overlapping_clusters.append(self.clusters[idx])

            if len(overlapping_clusters) == 1:
                return overlapping_clusters[0]
            else:
                raise Exception(f'More than one clusters overlap with input range. Overlapping clusters: {overlapping_clusters}')
        

class FetchedReads:
    def __init__(self, bam, readfilter=None, addition_padding=DEFAULT_FETCHEDREADS_ADDITION_PADDING):
        self.bam = bam
        self.addition_padding = addition_padding

        if readfilter is None:
            self.readfilter = readhandler.readfilter_pileup
        else:
            self.readfilter = readfilter

        self.subdata = dict()

    def __repr__(self):
        buffer = list()
        buffer.append(f'<FetchedReads object, which consists of:')
        for chrom, frsinglechrom in self.subdata.items():
            buffer.append(f'\t{frsinglechrom}')
        buffer.append(f'>')
        return '\n'.join(buffer)

    @classmethod
    def from_fetch(cls, bam, chrom, start0, end0, readfilter=None, addition_padding=DEFAULT_FETCHEDREADS_ADDITION_PADDING):
        result = cls(bam, readfilter, addition_padding)
        result.subdata[chrom] = FetchedReadsSinglechrom.from_fetch(result.bam, chrom, start0, end0, result.readfilter, result.addition_padding)

        return result

    def get_read(self, read_uid):
        found_read = False
        for frcluster in itertools.chain.from_iterable(
            (cluster for cluster in frsinglechr.clusters)
            for frsinglechr in self.subdata.values()
        ):
            try:
                read = frcluster.dict[read_uid]
            except KeyError:
                continue
            else:
                found_read = True
                break

        if not found_read:
            raise Exception(f'Given ReadUID is not present.')

        return read

    def iter_reads(self, chrom=None):
        if chrom is None:
            return itertools.chain.from_iterable(x.iter_reads() for x in self.subdata.values())
        else:
            return self.subdata[chrom].iter_reads()

    def add_reads(self, chrom, start0, end0):
        if chrom not in self.subdata:
            self.subdata[chrom] = FetchedReadsSinglechrom(chrom, self.bam, self.readfilter, self.addition_padding)
        self.subdata[chrom].add_reads(start0, end0)

    def fetch(self, chrom, start0, end0, with_uid=False):
        if chrom not in self.subdata:
            self.subdata[chrom] = FetchedReadsSinglechrom(chrom, self.bam, self.readfilter, self.addition_padding)
        return self.subdata[chrom].fetch(start0, end0, with_uid=with_uid)

    def naive_fetch(self, chrom, start0, end0, with_uid=False):
        if chrom not in self.subdata:
            raise Exception(f'Input chrom is not included in the database.')
        return self.subdata[chrom].naive_fetch(start0, end0, with_uid=with_uid)

    def check_fully_includes(self, chrom, start0, end0):
        if chrom not in self.subdata:
            return False
        else:
            return self.subdata[chrom].check_fully_includes(start0, end0)

    def choose_cluster(self, chrom, start0, end0):
        if chrom not in self.subdata:
            return False
        else:
            return self.subdata[chrom].choose_cluster(start0, end0)


####################################


class FetchedRefseqCluster(ClusterBase):
    def __init__(self, fasta, chrom):
        self.fasta = fasta
        self.chrom = chrom
        self.start0 = None
        self.end0 = None
        self.seq = None

    def __repr__(self):
        return f'<FetchedRefseqCluster object (chrom:{self.chrom}, start0:{self.start0}, end0:{self.end0})>'

    @classmethod
    def from_fetch(cls, fasta, chrom, start0, end0):
        result = cls(fasta, chrom)
        result.start0 = start0
        result.end0 = end0
        result.seq = list(fasta.fetch(chrom, start0, end0))
        print('fasta fetch')

        return result

    @property
    def range0(self):
        return range(self.start0, self.end0)

    def get_id(self):
        return (self.chrom, self.start0, self.end0)

    def copy(self):
        result = self.__class__(self.fasta, self.chrom)
        result.start0 = self.start0
        result.end0 = self.end0
        result.seq = self.seq.copy()
        return result

    def fetch(self, start0, end0):
        start0, end0 = cluster_fetch_arghandler(start0, end0, self.start0, self.end0)
        start_idx = start0 - self.start0
        end_idx = end0 - self.start0
        return ''.join(self.seq[start_idx:end_idx])

    def extend_rightward(self, width):
        self.seq.extend(self.fasta.fetch(self.chrom, self.end0, self.end0 + width))
        print('fasta fetch')
        self.end0 += width

    def extend_leftward(self, width):
        for base in reversed(self.fasta.fetch(self.chrom, self.start0 - width, self.start0)):
            self.seq.insert(0, base)
        print('fasta fetch')
        self.start0 -= width

    def merge(self, other):
        if (other.start0 > self.end0) or (self.start0 > other.end0):
            raise Exception(f'Two objects do not overlap.')

        if self.start0 <= other.start0 and self.end0 >= other.end0:
            return self.copy()
        elif other.start0 <= self.start0 and other.end0 >= self.end0:
            return other.copy()
        else:
            # now self.start0 != other.start0
            if self.start0 < other.start0:
                left = self
                right = other
            else:
                left = other
                right = self

            result = self.__class__(self.fasta, self.chrom)
            result.start0 = left.start0
            result.end0 = right.end0
            result.seq = left.seq[:(right.start0 - left.start0)] + right.seq
            return result


class FetchedRefseqSinglechrom(SinglechromBase):
    def __init__(self, chrom, fasta, addition_padding=DEFAULT_FETCHEDREFSEQ_ADDITION_PADDING):
        self.chrom = chrom
        self.fasta = fasta
        self.addition_padding = addition_padding
        self.clusters = list()

    def __repr__(self):
        return f'<FetchedRefseqSinglechrom object (chrom={self.chrom}, clusters={self.clusters})>'

    @classmethod
    def from_fetch(cls, fasta, chrom, start0, end0, addition_padding=DEFAULT_FETCHEDREFSEQ_ADDITION_PADDING):
        result = cls(chrom, fasta, addition_padding)
        result.clusters.append(FetchedRefseqCluster.from_fetch(result.fasta, result.chrom, start0, end0))
        return result

    def spawn_cluster_from_fetch(self, start0, end0):
        return FetchedRefseqCluster.from_fetch(self.fasta, self.chrom, start0, end0)

    def add_seq(self, start0, end0):
        self.add(start0, end0)

    def fetch(self, start0, end0):
        valid_cluster = self.get_valid_cluster(start0, end0)
        return valid_cluster.fetch(start0, end0)


class FetchedRefseq:
    def __init__(self, fasta, addition_padding=DEFAULT_FETCHEDREFSEQ_ADDITION_PADDING):
        self.fasta = fasta
        self.addition_padding = addition_padding
        self.subdata = dict()

    def __repr__(self):
        buffer = list()
        buffer.append(f'<FetchedRefseq object, which consists of:')
        for chrom, frsinglechrom in self.subdata.items():
            buffer.append(f'\t{frsinglechrom}')
        buffer.append(f'>')
        return '\n'.join(buffer)

    def fetch(self, chrom, start0, end0):
        if chrom not in self.subdata:
            self.subdata[chrom] = FetchedRefseqSinglechrom(chrom, self.fasta, self.addition_padding)
        return self.subdata[chrom].fetch(start0, end0)


###########################





import sys
import collections
import itertools
import functools
import logging
import inspect
import random
import uuid

import pysam
import Bio.Align
import numpy as np
import pandas as pd
import pyranges as pr

import handygenome.tools as tools
import handygenome.workflow as workflow
import handygenome.variant.vcfspec as libvcfspec
import handygenome.read.pileup as libpileup
import handygenome.align.alignhandler as alignhandler
import handygenome.bameditor as bameditor
import handygenome.read.readhandler as readhandler
import handygenome.read.readplus as readplus
import handygenome.align.realign.base as realign_base
import handygenome.align.realign.msrpileup as realign_msrpileup
import handygenome.align.realign.rpileup_series as realign_rpileup_series


class MultisampleRealignerPileupSeries:
    def __repr__(self):
        return str(self.pileupseries_dict)

    def iter_series(self):
        return iter(self.pileupseries_dict.values())

    def iter_mspileups(self):
        return iter(self.mspileup_list)

    @functools.cached_property
    def df(self):
        data = list()
        index = list()
        for idx, (sampleid, rpileup_series) in enumerate(self.pileupseries_dict.items()):
            index.append(sampleid)
            data.append(list(rpileup_series.pileup_list))
            if idx == 0:
                breaks = list()
                breaks.append(rpileup_series.pileup_list[0].start0)
                for rpileup in rpileup_series.pileup_list:
                    breaks.append(rpileup.end0)
                columns = pd.IntervalIndex.from_breaks(breaks, closed='left')
        return pd.DataFrame.from_records(data, index=index, columns=columns)

#    @property
#    def rows(self):
#        return pd.DataFrame.from_dict(self.pileupseries_dict, orient='index')
#
#    @property
#    def cols(self):
#        return pd.DataFrame.from_records(
#            [
#                {
#                    (x.start0, x.end0): x 
#                    for x in self.mspileup_list
#                }
#            ]
#        )

    #@property
    #def series_length(self):
        #return len(self.rows.iloc[0, :])
    #    return len(next(iter(self.pileupseries_dict.values())))

    @property
    def ranges(self):
        #return [range(*x) for x in self.cols.columns]
        return [x.range0 for x in next(iter(self.pileupseries_dict.values())).pileup_list]

    @property
    def start0(self):
        #return self.rows.iloc[0, :].start0
        return next(self.pileupseries_dict.values()).start0

    @property
    def end0(self):
        #return self.rows.iloc[0, :].end0
        return next(self.pileupseries_dict.values()).end0

    def write_realigned_reads(self, bam_path_dict, padding=0):
        for sampleid, rpileup_ser in self.pileupseries_dict.items():
            rpileup_ser.write_realigned_reads(bam_path_dict[sampleid], padding=padding)

    def iter_contig_vcfspecs(self, subseq_portion_threshold=0, MQ_threshold=40):
        return itertools.chain.from_iterable(
            msrpileup.iter_contig_vcfspecs(subseq_portion_threshold=subseq_portion_threshold, MQ_threshold=MQ_threshold)
            for msrpileup in self.mspileup_list
        )

    def get_result_vcfspecs(self, portion_cutoff=0.1, MQ_cutoff=30, verbose=False):
        return list(
            msrpileup.get_result_vcfspec(
                subseq_portion_threshold=portion_cutoff, 
                MQ_threshold=MQ_cutoff, 
                threshold_bysample=True,
                verbose=verbose,
            )
            for msrpileup in self.mspileup_list
        )

    def get_result_vcfspecs_old(
        self, vaf_cutoff=0.01, subseq_portion_threshold=None, 
        MQ_threshold=40, threshold_bysample=True,
        verbose=False,
    ):
        return list(
            msrpileup.get_result_vcfspec_old(
                vaf_cutoff=vaf_cutoff, 
                subseq_portion_threshold=subseq_portion_threshold, 
                MQ_threshold=MQ_threshold, 
                threshold_bysample=threshold_bysample,
                verbose=verbose,
            )
            for msrpileup in self.mspileup_list
        )

    def get_result_vcfspecs_old(
        self, as_components=False, merge=True, subseq_portion_threshold=None, MQ_threshold=40, threshold_bysample=True,
    ): 
        return list(
            msrpileup.get_result_vcfspecs(
                as_components=as_components, 
                merge=merge, 
                subseq_portion_threshold=subseq_portion_threshold, 
                MQ_threshold=MQ_threshold, 
                threshold_bysample=threshold_bysample,
            )
            for msrpileup in self.mspileup_list
        )

    ###########################
    # initializer and helpers #
    ###########################
    def __init__(
        self, 
        bam_dict, 
        chrom, 
        start0, 
        end0, 
        refver=None,
        fasta=None,
        aligner=realign_base.DEFAULT_ALIGNER,
        verbose=False, 
        logger=None, 
        init_blank=False,
        **kwargs,
    ):
        # set params
        self.chrom = chrom
        self.bam_dict = bam_dict
        self.refver, self.fasta = realign_base.RealignerPileupBase.refver_fasta_arghandler(refver, fasta)
        self.aligner = aligner
        self.verbose, self.logger = realign_base.RealignerPileupBase.logger_arghandler(verbose, logger)
        self.params = realign_base.parse_rpileup_kwargs(**kwargs)

        # setup data
        if not init_blank:
            self.set_series_dict(start0, end0)  # self.pileupseries_dict
            self.set_multisample_pileups()  # self.mspileup_list
            self.set_realigned_reads()

    def set_series_dict(self, seed_start0, seed_end0):
        self.pileupseries_dict = dict()
        self.no_variant = False

        # initialize
        for sampleid, bam in self.bam_dict.items():
            self.logger.debug(f'@@@ Initializing RealignerPileupSeries of sample {sampleid} @@@\n')
            self.pileupseries_dict[sampleid] = realign_rpileup_series.RealignerPileupSeries(
                bam=bam, 
                chrom=self.chrom, start0=seed_start0, end0=seed_end0, 
                refver=self.refver,
                fasta=self.fasta, 
                aligner=self.aligner,
                verbose=self.verbose,
                logger=self.logger,
                **self.params,
            )
            self.logger.debug(f'@@@ Finished initialization of RealignerPileupSeries of sample {sampleid} @@@\n\n')

        # equalize whole margins
        self.equalize_left()
        self.equalize_right()

        # equalize sub-pileup margins
        self.logger.debug(f'@@@ Beginning inner margin equalization @@@\n')
        self.equalize_inner_margins()
        self.logger.debug(f'@@@ Finished inner margin equalization @@@\n')

    def equalize_left(self):
        self.logger.debug(f'@@@ Beginning equalize_left @@@\n')

        # equalize left
        while True:
            if len(set(x.start0 for x in self.pileupseries_dict.values())) == 1:
                break
            # extend
            target_start0 = min(x.start0 for x in self.pileupseries_dict.values())
            self.logger.debug(f'target_start0: {target_start0}')
            for sampleid, pileup_ser in self.pileupseries_dict.items():
                width = pileup_ser.start0 - target_start0
                if width > 0:
                    self.logger.debug(f'Beginning EXTEND of {sampleid}')
                    self.logger.debug(f'current start0 of {sampleid}: {pileup_ser.start0}')
                    pileup_ser.extend_left(width)
                    self.logger.debug(f'Finished EXTEND of {sampleid}\n')

            #for sampleid, pileup_ser in self.pileupseries_dict.items():
                #print(sampleid)
                #print(pileup_ser.pileup_list[0].df)
            #    print(pileup_ser.pileup_list[0].active_info)
            #    print()
            #print('----------------')

            # check if hit width limit
            if any(
                pileup_ser.width >= self.params['max_series_width']
                for pileup_ser in self.pileupseries_dict.values()
            ):
                break
            # secure
            for sampleid, pileup_ser in self.pileupseries_dict.items():
                self.logger.debug(f'Beginning SECURE of {sampleid}')
                pileup_ser.secure_left()
                self.logger.debug(f'Finished SECURE of {sampleid}\n')

        self.logger.debug(f'@@@ Finished equalize_left @@@\n')

    def equalize_right(self):
        self.logger.debug(f'@@@ Beginning equalize_right @@@\n')

        while True:
            if len(set(x.end0 for x in self.pileupseries_dict.values())) == 1:
                break
            # extend
            target_end0 = max(x.end0 for x in self.pileupseries_dict.values())
            self.logger.debug(f'target_end0: {target_end0}')
            for sampleid, pileup_ser in self.pileupseries_dict.items():
                width = target_end0 - pileup_ser.end0
                if width > 0:
                    self.logger.debug(f'Beginning EXTEND of {sampleid}')
                    self.logger.debug(f'current end0 of {sampleid}: {pileup_ser.end0}')
                    pileup_ser.extend_right(width)
                    self.logger.debug(f'Finished EXTEND of {sampleid}\n')

            for sampleid, pileup_ser in self.pileupseries_dict.items():
                print(sampleid)
                #print(pileup_ser.pileup_list[-1].df)
                print(pileup_ser.pileup_list[-1].active_info)
                print()
            print('----------------')

            # check if hit width limit
            if any(
                pileup_ser.width >= self.params['max_series_width'] 
                for pileup_ser in self.pileupseries_dict.values()
            ):
                break
            # secure
            for pileup_ser in self.pileupseries_dict.values():
                self.logger.debug(f'Beginning SECURE of {sampleid}')
                pileup_ser.secure_right()
                self.logger.debug(f'Finished SECURE of {sampleid}\n')

        self.logger.debug(f'@@@ Finished equalize_right @@@\n')

    def equalize_inner_margins(self):
        # set interim parameters
        first_pileupseries = next(self.iter_series())
        series_start0 = first_pileupseries.start0
        series_end0 = first_pileupseries.end0
        series_width = series_end0 - series_start0
        max_pileup_width = first_pileupseries.pileup_list[0].params['max_pileup_width']

        # When there is no need to further split the series range
        if series_end0 - series_start0 <= max_pileup_width:
            start0_list = [series_start0, series_end0]
            for pileup_ser in self.pileupseries_dict.values():
                if len(pileup_ser.pileup_list) > 1:
                    pileup_ser.rearrange(start0_list, prepare_vcfspecs=True)
            self.rearrange_mode = 'not_done'
            return

        # best case - without splitting contig vcfspec
        start0_list, all_splittable = self.get_rearrangement_points(
            split_region_gr=functools.reduce(
                lambda x, y: x.intersect(y), 
                (pileup_ser.get_splittable_region_best() for pileup_ser in self.pileupseries_dict.values())
            ), 
            trim_margins=True, 
            max_pileup_width=max_pileup_width, 
            series_start0=series_start0, 
            series_end0=series_end0,
        )

        if start0_list is not None:
            for pileup_ser in self.pileupseries_dict.values():
                pileup_ser.rearrange(start0_list, prepare_vcfspecs=True)
            self.rearrange_mode = 'preserve_contig'
            return
        else:
            if all_splittable:  # no variant case
                # split series range into evenly-spaced subsets
                start0_list = np.linspace(
                    series_start0,
                    series_end0,
                    np.ceil(series_width / max_pileup_width).astype('int') + 1,
                    endpoint=True,
                ).astype('int')
                for pileup_ser in self.pileupseries_dict.values():
                    pileup_ser.rearrange(start0_list, prepare_vcfspecs=True)

                self.no_variant = True
                self.rearrange_mode = None
                return

        # splitting contig vcfspec
        start0_list, all_splittable = self.get_rearrangement_points(
            split_region_gr=functools.reduce(
                lambda x, y: x.intersect(y), 
                (pileup_ser.get_splittable_region_split_contig() for pileup_ser in self.pileupseries_dict.values())
            ), 
            trim_margins=True, 
            max_pileup_width=max_pileup_width, 
            series_start0=series_start0, 
            series_end0=series_end0,
        )

        if start0_list is not None:
            for pileup_ser in self.pileupseries_dict.values():
                pileup_ser.rearrange(start0_list, prepare_vcfspecs=True)
            self.rearrange_mode = 'split_contig'
            return
        else:
            if all_splittable:
                raise Exception(f'all-splittable is True in split-contig after False in non-split-contig')

        # using inactive runs
        start0_list, all_splittable = self.get_rearrangement_points(
            split_region_gr=functools.reduce(
                lambda x, y: x.intersect(y), 
                (pileup_ser.get_splittable_region_inactive_runs() for pileup_ser in self.pileupseries_dict.values())
            ), 
            trim_margins=False, 
            max_pileup_width=max_pileup_width, 
            series_start0=series_start0, 
            series_end0=series_end0,
        )

        if start0_list is not None:
            for pileup_ser in self.pileupseries_dict.values():
                pileup_ser.rearrange(start0_list, prepare_vcfspecs=True)
            self.rearrange_mode = 'inactive_runs'
            return
        else:
            if all_splittable:
                raise Exception(f'All PileupSeries region is inactive.')
            else:
                self.rearrange_mode = 'failed'
                return

    @staticmethod
    def get_rearrangement_points(split_region_gr, trim_margins, max_pileup_width, series_start0, series_end0):
        """Returns:
            start0_list: list of 0-based positions. 
                With pairwise iteration, for each iteratation result (x, y),
                x is used as start0 and y as end0 for a RealignerPileup object.
        """
        if split_region_gr.empty:
            # abort this split strategy
            start0_list = None
            all_splittable = False
            return start0_list, all_splittable

        # when entire range is splittable
        if split_region_gr.df.shape[0] == 1:
            row = split_region_gr.df.iloc[0, :]
            if (row.Start == series_start0) and (row.End == series_end0):
                start0_list = None
                all_splittable = True
                return start0_list, all_splittable

        # make into ranges
        start_candidate_ranges = [range(row.Start, row.End + 1) for (idx, row) in split_region_gr.df.iterrows()]

        # trim or add margins
        if start_candidate_ranges[0].start == series_start0:
            if trim_margins:
                start_candidate_ranges[0] = range(start_candidate_ranges[0].stop - 1, start_candidate_ranges[0].stop)
        else:
            start_candidate_ranges.insert(0, range(series_start0, series_start0 + 1))

        if start_candidate_ranges[-1].stop - 1 == series_end0:
            if trim_margins:
                start_candidate_ranges[-1] = range(start_candidate_ranges[-1].start, start_candidate_ranges[-1].start + 1)
        else:
            start_candidate_ranges.append(range(series_end0, series_end0 + 1))

        # now len(start_candidate_ranges) is at least 2 

        # If candidate range is smaller than max_pileup_width after trimming
        if trim_margins:
            new_start0 = start_candidate_ranges[0].start
            new_end0 = start_candidate_ranges[-1].stop - 1
            if new_end0 - new_start0 <= max_pileup_width:
                start0_list = [new_start0, new_end0]
                all_splittable = False
                return start0_list, all_splittable

        # validity check
        if any(
            rng2.start - (rng1.stop - 1) > max_pileup_width
            for rng1, rng2 in tools.pairwise(start_candidate_ranges)
        ):
            # abort this split strategy
            start0_list = None
            all_splittable = False
            return start0_list, all_splittable

        # search for actual split points
        iterator = iter(start_candidate_ranges)
        current_rng = next(iterator)
        next_rng = next(iterator)
        start0_list = [current_rng.start]
        stopiter = False

        while True:
            if stopiter:
                break

            last_start0 = start0_list[-1]
            while True:
                if next_rng.start <= last_start0 + max_pileup_width:
                    current_rng = next_rng

                    try:
                        next_rng = next(iterator)
                    except StopIteration:
                        stopiter = True
                        break
                else:
                    if current_rng.stop - 1 <= last_start0:
                        raise Exception(f'Cannot make pileup margins within max_pileup_width')
                    start0_list.append(
                        min(current_rng.stop - 1, last_start0 + max_pileup_width)
                    )
                    break

        start0_list.append(next_rng.start)

        all_splittable = False
        return start0_list, all_splittable

    def set_multisample_pileups(self):
        self.mspileup_list = list()
        num_col = len(next(self.iter_series()).pileup_list)
        for idx in range(num_col):
            pileup_dict = {
                sampleid: self.pileupseries_dict[sampleid].pileup_list[idx]
                for sampleid in self.pileupseries_dict.keys()
            }
            self.mspileup_list.append(
                realign_msrpileup.MultisampleRealignerPileup(
                    pileup_dict=pileup_dict, 
                    refver=self.refver,
                    fasta=self.fasta, 
                    aligner=self.aligner,
                    verbose=self.verbose, 
                    logger=self.logger,
                    **self.params,
                )
            )

    def set_realigned_reads(self):
        """Assumes all mspileups have executed 'save_subseq_alignments' and 'allocate_subseq_alignments'"""
        #for msrpileup in self.mspileup_list:
        #    msrpileup.save_subseq_alignments()
        #    msrpileup.allocate_subseq_alignments()
        self.realigned_reads = {sampleid: dict() for sampleid in self.pileupseries_dict.keys()}
        for sampleid, rpileup_ser in self.pileupseries_dict.items():
            rpileup_ser.set_realigned_reads()
            self.realigned_reads[sampleid] = rpileup_ser.realigned_reads



import json
import re
import contextlib
import time
import shutil
import subprocess

import ftplib
import urllib.request
import urllib.error
import urllib.parse

import handygenome.logutils as logutils
import handygenome.deco as deco


HTTP_HEADER_POST = {
    "Content-Type": "application/json", 
    "Accept": "application/json",
}
HTTP_HEADER_GET = {'Content-Type': 'application/json'}


class MaxRetryError(Exception):
    pass


def retry_or_raise(n_try, retry_count, exc, retry_interval, msg_prefix=''):
    """helper function"""
    if (retry_count is None) or (n_try <= retry_count):
        retry_count_string = (
            'inf' if retry_count is None else str(retry_count)
        )
        logutils.log(
            f'{msg_prefix}Retrying due to failure (Exception: {exc}); n_try={n_try}/{retry_count_string}',
            level='info',
        )
#        if isinstance(exc, TimeoutError):
#            logutils.log(
#                f'{msg_prefix}Retrying due to TimeoutError (Exception: {exc}); '
#                f'n_try={n_try}/{retry_count_string}'
#            )
#        else:
#            logutils.log(
#                f'{msg_prefix}Retrying due to non-TimeoutError (Exception: {exc}); '
#                f'n_try={n_try}/{retry_count_string}'
#            )
        time.sleep(retry_interval)
    else:
        if isinstance(exc, urllib.error.URLError):
            print(exc.read().decode('utf-8'))
        else:
            print(str(exc))
        raise MaxRetryError(f'Exceeded maximum retry count({retry_count}).') from exc


# ftp

def ftp_login(url, retry_count=10, retry_interval=1, timeout=5):
    @deco.get_deco_timeout(timeout)
    def helper():
        ftp = ftplib.FTP(url, timeout=timeout)
        ftp.login()
        return ftp

    retry_msg_pf = 'ftp login: '

    n_try = 0
    while True:
        n_try += 1
        try:
            #with contextlib.redirect_stdout('/dev/null'):
            ftp = helper()
            #ftp = ftplib.FTP(url, timeout=timeout)
            #ftp.login()
        except TimeoutError as exc:
            retry_or_raise(n_try, retry_count, exc, retry_interval, msg_prefix=retry_msg_pf)
            continue
        except OSError as exc:
            if (str(exc) == '[Errno 101] Network is unreachable'):
                retry_or_raise(n_try, retry_count, exc, retry_interval, msg_prefix=retry_msg_pf)
                continue
            else:
                raise
        else:
            break

    return ftp



def ftp_nlst(
    authority, path, 
    retry_count=10, retry_interval=1, timeout=5, verbose=False,
):
    retry_msg_pf = 'ftp nlst: '

    if verbose:
        logutils.log(
            f'Performing ftp nlst (authority={repr(authority)}, path={repr(path)})',
            level='debug',
        )

    n_try = 0
    while True:
        n_try += 1
        try:
            ftp = ftp_login(
                authority, 
                retry_count=1, 
                retry_interval=retry_interval, 
                timeout=timeout,
            )
            result = ftp.nlst(path)
        except MaxRetryError as exc:  # ftp_login failure
            retry_or_raise(n_try, retry_count, exc, retry_interval, msg_prefix=retry_msg_pf)
            continue
        except TimeoutError as exc:  # nlst failure
            retry_or_raise(n_try, retry_count, exc, retry_interval, msg_prefix=retry_msg_pf)
            continue
#        except OSError as exc:  # nlst failure when ftp object is timed-out
#            if (str(exc) == 'cannot read from timed out object'):
#                retry_or_raise(n_try, retry_count, exc, retry_interval, msg_prefix=retry_msg_pf)
#                continue
#            else:
#                raise
        else:
            break

    return result


def trim_path(path):
    path = re.sub('/+$', '', path)
    path = re.sub('/{2,}', '/', path)
    return path


def join_ftp_paths(*args):
    return trim_path('/'.join(args))


def ftp_listdir(ftp, path):
    return ftp.nlst(trim_path(path))


def ftp_listdir_old(ftp, path):
    path = trim_path(path)
    fname_list = list()
    ftp.cwd(path)
    ftp.retrlines('NLST', callback=(lambda x: fname_list.append(path + '/' + x)))
    return fname_list


# http

def http_run_urlopen(url_or_req, retry_count=10, retry_interval=1, urlopen_timeout=5):
    url_string = (
        url_or_req
        if isinstance(url_or_req, str) else
        url_or_req.full_url
    )
    logutils.log(f'Trying to open url {repr(url_string)}', level='info')

    n_try = 0
    while True:
        n_try += 1
        try:
            response = urllib.request.urlopen(url_or_req, timeout=urlopen_timeout)
        except urllib.error.URLError as exc:
            if isinstance(exc, urllib.error.HTTPError):
                if exc.code == 500:  # internal server error
                    retry_or_raise(n_try, retry_count, exc, retry_interval)
                    continue
                else:
                    print(exc.read().decode('utf-8'))
                    raise
            else: 
                if str(exc) == '<urlopen error [Errno 101] Network is unreachable>':
                    retry_or_raise(n_try, retry_count, exc, retry_interval)
                    continue
                else:
                    print(exc.read().decode('utf-8'))
                    raise
        except TimeoutError as exc:
            retry_or_raise(n_try, retry_count, exc, retry_interval)
            continue
        else:
            break

    logutils.log(f'Succeeded to open url {repr(url_string)}', level='info')

    return response


def http_get(url, params=None, headers=None, text=False, retry_count=10, retry_interval=1):
    # set params
    if params is not None:
        url = url + '?' + urllib.parse.urlencode(params)
    if headers is None:
        if text:
            headers = {'Accept': 'text/plain'}
        else:
            headers = {'Accept': 'application/json'}
    # main
    req = urllib.request.Request(url, headers=headers, method='GET')
    return http_send_request(req, text, retry_count, retry_interval)


def http_post(url, data, params=None, headers=None, text=False, retry_count=10, retry_interval=1):
    # set params
    data = json.dumps(data).encode('ascii')
    if params is not None:
        url = url + '?' + urllib.parse.urlencode(params)
    if headers is None:
        if text:
            headers = {'Accept': 'text/plain'}
        else:
            headers = {'Accept': 'application/json'}
    # main
    req = urllib.request.Request(url, data=data, headers=headers, method='POST')
    return http_send_request(req, text, retry_count, retry_interval)


def http_send_request(req, text, retry_count=10, retry_interval=1, urlopen_timeout=5):
    with http_run_urlopen(
        req, 
        retry_count=retry_count, 
        retry_interval=retry_interval, 
        urlopen_timeout=urlopen_timeout,
    ) as response:
        if text:
            result = response.read().decode('utf-8')
        else:
            result = json.loads(response.read())

    return result


def download(url, path, retry_count=10, retry_interval=1, urlopen_timeout=5):
    logutils.log(f'Beginning download: url={repr(url)}, path={repr(path)}', level='info')
    while True:
        try:
            with http_run_urlopen(
                url, 
                retry_count=retry_count, 
                retry_interval=retry_interval, 
                urlopen_timeout=urlopen_timeout,
            ) as response:
                with open(path, 'wb') as outfile:
                    shutil.copyfileobj(response, outfile)
        except TimeoutError as exc:
            logutils.log(f'Retrying due to TimeoutError (Exception: {exc})', level='info')
            continue
        else:
            break
    logutils.log(f'Finished download: url={repr(url)}, path={repr(path)}', level='info')


def download_wget(url, path):
    subprocess.run(['wget', '-O', path, url])




import collections

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

import handygenome.workflow as workflow
import handygenome.variant.ponbams as libponbams
import handygenome.variant.filter as libfilter
import handygenome.deco as deco


def show_readcounts(
    readstats_dict,
    allele_index=1, 
    samples=None, 
    sample_label_colors=None,
    sort_by_value=False, 
    exclude_other=False,
    mask=None, 
    xlabel='sample ID',
    title=None,
    title_pad=None,

    bar_data_type='count',
    dot_data_type='vaf',

    figsize=(20, 7),
    bar_ylim=None, 
    dot_ylim=(0, 1),
    logscale_count=False, 
    nan_value=0,
):
    """Args:
        samples: A sequence of sample IDs, which determines plotting order
        sort_by_value: If True, samples are further sorted by ALT rppcount value, in increasing order
        mask: A dictionary (keys: sample IDs, values: True or False); True bars are colored blue, False bars are red
    """
    # samples argument sanity check
    if samples is not None:
        if not set(samples).issubset(readstats_dict.keys()):
            raise Exception(f'"samples" must be a subset of the keys of "readstats_dict".')

    # handle samples argument
    if samples is None:
        samples = list()
        for x in sorted(readstats_dict.keys()):
            if not readstats_dict[x].is_missing:
                samples.append(x)
    if sort_by_value:
        samples = sorted(
            samples, 
            key=(lambda x: readstats_dict[x]['rppcounts'][allele_index])
        )
    # handle sample_label_colors argument
    if sample_label_colors is None:
        sample_label_colors = np.repeat('black', len(samples))

    # plot parameters
    col_bar = 'tab:blue'
    col_bar_label = col_bar #'tab:blue'
    col_bar_true = col_bar #'tab:blue'
    col_bar_false = 'tab:red'

    col_dot = 'tab:purple'
    col_dot_label = col_dot #'tab:red'
    col_dot_nan = 'tab:red'

    # prepare mask
    if mask is None:
        mask = {sid: True for sid in samples}
    not_mask = {key: (not val) for key, val in mask.items()}

    mask_array = np.array([mask[key] for key in samples])
    not_mask_array = np.array([not_mask[key] for key in samples])

    # prepare data
    target_rpp_counts = np.array(
        [
            readstats_dict[sid]['rppcounts'][allele_index]
            for sid in samples
        ]
    )
    total_rpp_counts = np.array(
        [
            readstats_dict[sid].get_total_rppcount(exclude_other=exclude_other)
            for sid in samples
        ]
    )
    nontarget_rpp_counts = total_rpp_counts - target_rpp_counts
    vafs = np.array(
        [
            readstats_dict[sampleid].get_vaf(
                alleleclass=allele_index, 
                exclude_other=exclude_other,
            )
            for sampleid in samples
        ]
    )
    #vafs[np.isnan(vafs)] = nan_value  # nan is substituted with a negative value

    y_candidates = {
        'count': {
            'value': target_rpp_counts, 
            'label': 'ALT read count',
        },
        'vaf': {
            'value': vafs, 
            'label': 'vaf',
        },
    }
    
    # prepare x, y axis values
    bar_data_type = 'count'
    dot_data_type = 'vaf'

    x = np.arange(len(samples))

    #y_bar = y_candidates[bar_data_type]['value']
    #y_bar_label = y_candidates[bar_data_type]['label']
    y_bar_target = target_rpp_counts
    y_bar_nontarget = nontarget_rpp_counts
    y_bar_label = f'Read pair count (BAR)\n(counts with relevant allele index is colored)'

    #y_dot = y_candidates[dot_data_type]['value']
    #y_dot_label = y_candidates[dot_data_type]['label']
    y_dot = vafs
    y_dot_label = f'VAF (SCATTER)'

    # init figure
    fig, ax_bar = plt.subplots(figsize=figsize)

    # ax_bar
    ax_bar.set(xticks=np.arange(len(samples)), xticklabels=list(samples))
    for label, color in zip(ax_bar.get_xticklabels(), sample_label_colors):
        label.set_color(color)

    # true bars
    width = 0.6
    ax_bar.bar(x[mask_array], y_bar_nontarget[mask_array], edgecolor='black', linewidth=1, fill=False, alpha=0.7, width=width)
    ax_bar.bar(x[mask_array], y_bar_target[mask_array], bottom=y_bar_nontarget[mask_array], edgecolor='black', linewidth=1, color=col_bar_true, alpha=0.7, width=width)
    # false bars
    ax_bar.bar(x[not_mask_array], y_bar_nontarget[not_mask_array], edgecolor='black', linewidth=1, fill=False, alpha=0.7, width=width)
    ax_bar.bar(x[not_mask_array], y_bar_target[not_mask_array], bottom=y_bar_nontarget[not_mask_array], edgecolor='black', linewidth=1, color=col_bar_false, alpha=0.7, width=width)
    #ax_bar.bar(x[not_mask_array], y_bar[not_mask_array], color=col_bar_false, alpha=0.7)

    ax_bar.set_xlabel(xlabel, size=15)
    ax_bar.tick_params(axis='x', direction='out', labelrotation=90)
    #ax_bar.set_ylabel(f'{y_bar_label} (BAR)', color=col_bar_label)
    ax_bar.set_ylabel(y_bar_label, color=col_bar_label)
    ax_bar.tick_params(axis='y', color=col_bar_label, labelcolor=col_bar_label)

    if title is not None:
        ax_bar.set_title(title, pad=title_pad)

    if bar_ylim is not None:
        ax_bar.set(ylim=bar_ylim)
    else:
        ax_bar.set(
            ylim=(0, np.ceil(max(total_rpp_counts)))
        )

    if logscale_count:
        ax_bar.set_yscale('log')
        
    # ax_dot
    ax_dot = ax_bar.twinx()
    y_isnan = np.isnan(y_dot)
    y_for_nan = np.repeat(nan_value, y_isnan.sum())

    markersize = 8
    ax_dot.plot(
        x[~y_isnan], 
        y_dot[~y_isnan], 
        color=col_dot,
        marker='o',
        markersize=markersize,
        linestyle='',
    )
    ax_dot.plot(
        x[y_isnan], 
        y_for_nan,
        color=col_dot_nan,
        marker='x',
        markersize=markersize,
        linestyle='',
    )

    if dot_ylim is not None:
        ax_dot.set(ylim=dot_ylim)
    else:
        ax_dot.set(ylim=(-0.03, max(y_dot) * 1.2))

    #ax_dot.set_ylabel(f'{y_dot_label} (SCATTER)', color=col_dot_label)
    ax_dot.set_ylabel(y_dot_label, color=col_dot_label)
    ax_dot.tick_params(axis='y', color=col_dot_label, labelcolor=col_dot_label)

    return fig, ax_bar, ax_dot


@deco.get_deco_num_set_differently(('pon_samples', 'pon_cohorts'), 1)
def show_pon(
    query_sample, 
    vp, 
    allele_index=1, 
    exclude_query=True, 
    exclude_other=False,
    pon_cohorts=None, 
    pon_samples=None, 
    ponfilter_mode='wgs',
    zoom_lower=False,
    **kwargs,
):
    assert ponfilter_mode in ('wgs', 'panel_germline', 'panel_somatic')
    assert (
        (pon_cohorts is None and pon_samples is not None) or 
        (pon_cohorts is not None and pon_samples is None)
    )
    assert isinstance(pon_cohorts, (list, tuple, type(None))), f'"pon_cohorts" must be a list or a tuple.'
    assert isinstance(pon_samples, (list, tuple, type(None))), f'"pon_samples" must be a list or a tuple.'

    # set pon_samples from cohort name
    if pon_cohorts is not None:
        pon_samples = libponbams.get_pon_sample_names(pon_cohorts, vp.refver)
    else:
        pon_samples = list(pon_samples)

    # remove query sample from pon samples
    if exclude_query:
        if query_sample in pon_samples:
            pon_samples.remove(query_sample)

    # sanity check
    if query_sample not in vp.readstats_dict.keys():
        raise Exception(f'"query_sample" is not included in readstats_dict')
    if not set(pon_samples).issubset(vp.readstats_dict.keys()):
        raise Exception(f'"pon_samples" is not included in readstats_dict')

    # main
    # run ponfilter
    if ponfilter_mode == 'wgs':
        ponfilter = libfilter.PonFilterWGS(samples=pon_samples, save_cache=True, verbose=False, **kwargs)
    elif ponfilter_mode == 'panel_germline':
        ponfilter = libfilter.PonFilterPanelseqGermline(samples=pon_samples, save_cache=True, verbose=False, **kwargs)
    elif ponfilter_mode == 'panel_somatic':
        ponfilter = libfilter.PonFilterPanelseqSomatic(samples=pon_samples, save_cache=True, verbose=False, **kwargs)

    ponfilter.check(vp, query_sample, allele_index=allele_index, exclude_query=exclude_query)
    
    # prepare params for show_readcounts
    title = '\n'.join([
        #f'{repr(vp.vcfspec).strip()}',
        f'{vp.vcfspec}',
        ' ; '.join([
            f'ponfilter_mode: {ponfilter_mode}',
            f'bisect_cutoff: {ponfilter.params["bisect_cutoff"]}',
            f'allele_index: {allele_index}',
        ]),
        ' ; '.join([
            f'lower_different: {ponfilter.cache["lower_different"]}',
            f'upper_different: {ponfilter.cache["upper_different"]}',
            f'lower_sample_fraction: {str(ponfilter.cache["lower_sample_fraction"])[:4]}',
            f'upper_sample_fraction: {str(ponfilter.cache["upper_sample_fraction"])[:4]}',
        ]),
        ' ; '.join([
            f'is_germline: {ponfilter.cache["is_germline"]}',
            f'het_vaf_mean: {str(ponfilter.cache["het_vaf_mean"])[:4]}',
            f'germline_fraction_among_upper: {str(ponfilter.cache["germline_fraction_among_upper"])[:4]}',
            f'germline_fraction_among_all: {str(ponfilter.cache["germline_fraction_among_all"])[:4]}',
        ]),
#        ' ; '.join([
#            f'is_global: {ponfilter.cache["is_global"]}',
#        ]),
        ' ; '.join([
            f'total_result: {ponfilter.cache["total_result"]}',
        ]),
    ])
    all_samples = [query_sample] + pon_samples
    sample_label_colors = [
        'red' if x in ponfilter.cache['samples_with_low_depth'] else 'black'
        for x in all_samples
    ]

    if ponfilter.cache['upper_cutoff'] is not None:
        dot_ylim = (0, max(ponfilter.cache['upper_cutoff'] + 0.1, 1))
    else:
        dot_ylim = (0, 1)

    fig, ax_bar, ax_dot = show_readcounts(
        vp.readstats_dict, 
        allele_index=allele_index,
        samples=all_samples,
        sample_label_colors=sample_label_colors,
        exclude_other=exclude_other,
        xlabel='sample ID (low-depth samples in red)',
        title=title,
        title_pad=30,
        dot_ylim=dot_ylim,
    )

    # Partition between query and PON samples 
    ax_bar.axvline(x=0.5, color='black')
    #offset = 0.15
    ax_bar.annotate(
        'query sample', 
        xy=(0, ax_bar.get_ylim()[1] * 1.01), 
        size=10,
        annotation_clip=False,
        ha='right',
        va='bottom',
        bbox=dict(boxstyle='round', facecolor='white', alpha=1),
    )
    ax_bar.annotate(
        'PON samples', 
        #xy=(0.5 + offset, ax_bar.get_ylim()[1] * 0.95), 
        xy=((1 + len(all_samples)) / 2, ax_bar.get_ylim()[1] * 1.01), 
        size=10,
        annotation_clip=False,
        ha='center',
        va='bottom',
        bbox=dict(boxstyle='round', facecolor='white', alpha=1),
    )

    # cutoff hlines
    ax_dot.axhline(y=ponfilter.params['bisect_cutoff'], color='black', linestyle='--')
    if ponfilter.cache['lower_mean'] is not None:
        ax_dot.axhline(y=ponfilter.cache['lower_mean'], color='orangered')
        ax_dot.axhline(y=ponfilter.cache['lower_cutoff'], color='lime')
    if ponfilter.cache['upper_mean'] is not None:
        ax_dot.axhline(y=ponfilter.cache['upper_mean'], color='orangered')
        ax_dot.axhline(y=ponfilter.cache['upper_cutoff'], color='lime')

    for text, y in [
        ('lower vafs mean', ponfilter.cache['lower_mean']),
        ('lower vafs cutoff', ponfilter.cache['lower_cutoff']),
        ('upper vafs mean', ponfilter.cache['upper_mean']),
        ('upper vafs cutoff', ponfilter.cache['upper_cutoff']),
    ]:
        if y is None:
            continue

        ax_dot.annotate(
            text, 
            xy=(ax_dot.get_xlim()[1] * 1.05, y), 
            size=10,
            annotation_clip=False,
            ha='left',
            va='center',
            bbox=dict(boxstyle='round', facecolor='white', alpha=1),
        )

    ax_dot_xlim = ax_dot.get_xlim()
    rectangles = list()
    if 'germline_het_range' in ponfilter.params:
        rectangles.append(
            mpl.patches.Rectangle(
                (ax_dot_xlim[0], ponfilter.params['germline_het_range'][0]),
                ax_dot_xlim[1] - ax_dot_xlim[0],
                ponfilter.params['germline_het_range'][1] - ponfilter.params['germline_het_range'][0],
            ),
        )
    if 'germline_hom_range' in ponfilter.params:
        rectangles.append(
            mpl.patches.Rectangle(
                (ax_dot_xlim[0], ponfilter.params['germline_hom_range'][0]),
                ax_dot_xlim[1] - ax_dot_xlim[0],
                ponfilter.params['germline_hom_range'][1] - ponfilter.params['germline_hom_range'][0],
            ),
        )
    if len(rectangles) > 0:
        ax_dot.add_collection(
            mpl.collections.PatchCollection(
                rectangles,
                alpha=0.1,
                facecolor='red',
                linewidth=0,
            )
        )

    if zoom_lower:
        ax_bar.set_ylim(0, ponfilter.params['bisect_cutoff'])
        ax_dot.set_ylim(0, ponfilter.params['bisect_cutoff'])

    return fig
    
    
###########################################

def show_readstats_data(readstats_data, alt_index=0, title=None, varpos_key='left'):
    fig, axs = plt.subplots(
        2, 6, 
        figsize=(40, 15),
        gridspec_kw={'hspace': 0.4},
    )
    #fig.tight_layout()
    
    # title
    if title is not None:
        fig.suptitle(title, size=30)
        
    # MQ
    plotter_MQ(ax=axs[0, 0], data=readstats_data['MQ'][alt_index + 1], 
               title='MQ - ALT reads')
    plotter_MQ(ax=axs[1, 0], data=readstats_data['MQ'][0], 
               title='MQ - REF reads')    
    # BQ
    plotter_BQ(ax=axs[0, 1], data=readstats_data['BQ'][alt_index + 1], 
               title='BQ - ALT reads')
    plotter_BQ(ax=axs[1, 1], data=readstats_data['BQ'][0], 
               title='BQ - REF reads')
    # pariorient
    plotter_pairorient(ax=axs[0, 2], data=readstats_data['pairorient'][alt_index + 1], 
                       title='pairorient - ALT reads')
    plotter_pairorient(ax=axs[1, 2], data=readstats_data['pairorient'][0], 
                       title='pairorient - REF reads')    
    # cliplen
    plotter_cliplen(ax=axs[0, 3], data=readstats_data['cliplen'][alt_index + 1], 
                    title='Softclip length - ALT reads')
    plotter_cliplen(ax=axs[1, 3], data=readstats_data['cliplen'][0], 
                    title='Softclip length - REF reads')
    # variant position
    fraction = varpos_key.endswith('_fraction')
    plotter_varpos(
        ax=axs[0, 4],
        data=readstats_data[f'pos0_{varpos_key}'][alt_index + 1],
        title=f'variant position from {varpos_key} end\nALT reads',
        fraction=fraction)
    plotter_varpos(
        ax=axs[1, 4],
        data=readstats_data[f'pos0_{varpos_key}'][0],
        title=f'variant position from {varpos_key} end\nREF reads',
        fraction=fraction)
    # read counts
    plotter_counts(ax=axs[0, 5], data=readstats_data['count'])
    plotter_counts_zoom(ax=axs[1, 5], data=readstats_data['count'])

    #plt.show()
    return fig

    
    
# show_readstats helpers
    
def plotter_hist(
    ax, data, title, xlim=None, binwidth=None, 
    fontsize_annot=18, fontsize_title=20,
    text_mean=True, text_median=True,
):
    ax.set_title(title, size=fontsize_title, pad=10)
    
    if xlim is not None:
        ax.set_xlim(xlim[0], xlim[1])
        
    if (xlim is not None) and (binwidth is not None):
        bins = np.arange(xlim[0], xlim[1] + binwidth, binwidth)    
    else:
        bins = None
    ax.hist(data, bins=bins)
        
    if text_mean:
        mean = np.mean(data)
        ax.text(
            0.05, 0.9, f'mean: {mean}', transform=ax.transAxes, 
            fontsize=fontsize_annot,
        )
    if text_median:
        median = np.median(data)
        ax.text(
            0.05, 0.8, f'median: {median}', transform=ax.transAxes, 
            fontsize=fontsize_annot,
        )
        
def plotter_bar(
    ax, data, title, indexes, ylim=None, text=None,
    fontsize_annot=18, fontsize_title=20,
    xlabelrotation=0,
):
    ax.set_title(title, size=fontsize_title, pad=10)
    ax.tick_params(axis='x', labelrotation=xlabelrotation)
    if ylim is not None:
        ax.set_ylim(ylim[0], ylim[1])
        
    ax.bar([str(x) for x in indexes], [data[x] for x in indexes])
    if text is not None:
        ax.text(0.05, 0.9, text, transform=ax.transAxes, fontsize=fontsize_annot)
        
###
        
def plotter_pairorient(ax, data, title):
    data_counter = collections.Counter(data)
    if data_counter['f2r1'] == 0:
        ratio = np.nan
    else:
        ratio = np.round(data_counter['f1r2'] / data_counter['f2r1'], 3)
    text = f'f1r2/f2r1 ratio: {ratio}'
    plotter_bar(ax, data_counter, title, indexes=['f1r2', 'f2r1'], text=text)
        
def plotter_MQ(ax, data, title):
    plotter_hist(ax, data, title, xlim=(0, 70), binwidth=1, 
                 text_mean=True, text_median=True)
    
def plotter_BQ(ax, data, title):
    plotter_hist(ax, data, title, xlim=(0, 40), binwidth=1, 
                 text_mean=True, text_median=True)
    
def plotter_cliplen(ax, data, title):
    plotter_hist(ax, data, title, xlim=(0, max(data) + 10), binwidth=None, 
                 text_mean=True, text_median=True)
    
def plotter_varpos(ax, data, title, fraction=False):
    if fraction:
        xlim = (0, 1)
        binwidth = 0.05
    else:
        xlim = (0, 155)
        binwidth = 5
    plotter_hist(ax, data, title, xlim=xlim, binwidth=binwidth, 
                 text_mean=True, text_median=True)
    
def plotter_counts(ax, data):
    data_counter = collections.Counter(data)
    total_count = get_total_rppcount(data_counter)
    data_counter['total'] = total_count
    title = 'Read counts per alleleclass'
    plotter_bar(ax, data_counter, title, 
                indexes=['total', 0, 1, -1, None, 'softclip_overlap'], 
                ylim=None, text=None, xlabelrotation=45)
    
def plotter_counts_zoom(ax, data):
    data_counter = collections.Counter(data)
    total_count = get_total_rppcount(data_counter)
    data_counter['total'] = total_count
    title = 'Read counts per alleleclass,\nzoom on ALT and softclip-overlaps'
    
    ylim_max = int(max(data_counter[1], data_counter['softclip_overlap']) * 2)
    ylim = (0, ylim_max)
    
    ratio = round(data_counter['softclip_overlap'] / data_counter[1], 3)
    text = f'softclip_overlap / ALT read count ratio: {ratio}'
    
    plotter_bar(ax, data_counter, title, 
                indexes=['total', 0, 1, -1, None, 'softclip_overlap'], 
                ylim=ylim, text=text, xlabelrotation=45)

def get_total_rppcount(rppcount_data_counter):
    return sum(val for (key, val) in rppcount_data_counter.items()
               if isinstance(key, int))




import sys
import collections
import itertools
import functools
import logging
import inspect
import random
import uuid

import pysam
import Bio.Align
import numpy as np
import pandas as pd
import pyranges as pr

import handygenome.tools as tools
import handygenome.workflow as workflow
import handygenome.variant.vcfspec as libvcfspec
import handygenome.read.pileup as libpileup
import handygenome.align.alignhandler as alignhandler
import handygenome.bameditor as bameditor
import handygenome.read.readhandler as readhandler
import handygenome.read.readplus as readplus
import handygenome.align.realign.base as realign_base
import handygenome.align.realign.rpileup as realign_rpileup


class RealignerPileupSeries:
    def __init__(
        self, 
        bam, 
        chrom, 
        start0, 
        end0, 
        refver=None, 
        fasta=None,
        aligner=realign_base.DEFAULT_ALIGNER,
        verbose=False,
        logger=None,
        **kwargs,
    ):
        """Args:
            **kwargs: keyword arguments of RealignerPileup.__init__
                except start0, end0, init_df
        """
        # set params
        self.chrom = chrom
        self.bam = bam
        self.refver, self.fasta = realign_base.RealignerPileupBase.refver_fasta_arghandler(refver, fasta)
        self.aligner = aligner
        self.verbose, self.logger = realign_base.RealignerPileupBase.logger_arghandler(verbose, logger)
        self.params = realign_base.parse_rpileup_kwargs(**kwargs)

        # setup pileup_list
        self._init_pileup_list(start0, end0)
        # merge individual read_store's
        self.read_store = dict()
        for pileup in self.pileup_list:
            self.read_store.update(pileup.read_store)

    def _init_pileup_list(self, start0, end0):
        # initiate pileup_list
        self.logger.debug(f'Initial pileup generation')

        initial_pileup, result_inactive, result_vcfspec = realign_rpileup.RealignerPileup.init_and_augment(
            bam=self.bam, 
            chrom=self.chrom, start0=start0, end0=end0, 
            refver=self.refver, fasta=self.fasta,
            aligner=self.aligner,
            verbose=self.verbose, 
            logger=self.logger, 
            **self.params,
        )
        self.pileup_list = [initial_pileup]

        self.logger.debug(f'Initial pileup: {initial_pileup}')
        self.logger.debug(f'result_inactive: {result_inactive}')
        self.logger.debug(f'result_vcfspec: {result_vcfspec}')
        self.logger.debug(f'Initial pileup generation Finished\n')

        if (result_inactive is None) and (result_vcfspec is None):
            # When there is no active region in the initial pileup
            return

        left_insufficient = (not result_inactive.left_okay) or (not result_vcfspec.left_okay)
        right_insufficient = (not result_inactive.right_okay) or (not result_vcfspec.right_okay)

        if left_insufficient:
            self.secure_left()
        if right_insufficient:
            self.secure_right()

        self.logger.debug(f'Finished initial pileup generation\n')

    def __repr__(self):
        string = '\n'.join(
            f'\t{idx} {pileup}'
            for idx, pileup in enumerate(self.pileup_list)
        )
        return (
            f'<'
            f'{self.__class__.__name__}'
            f'(pileup_list: [\n{string}\n])'
            f'>'
        )

    def __len__(self):
        return len(self.pileup_list)

    @property
    def start0(self):
        return self.pileup_list[0].start0

    @property
    def end0(self):
        return self.pileup_list[-1].end0

    @property
    def range0(self):
        return range(self.start0, self.end0)

    @property
    def width(self):
        return self.pileup_list[-1].end0 - self.pileup_list[0].start0

    def get_gr(self):
        start0s = list()
        end0s = list()
        for pileup in self.pileup_list:
            start0s.append(pileup.start0)
            end0s.append(pileup.end0)
        chroms = [self.chrom] * len(self.pileup_list)
        self_indexes = list(range(len(self.pileup_list)))
        return pr.from_dict(
            {'Chromosome': chroms, 'Start': start0s, 'End': end0s, 'Self_index': self_indexes}
        )

    def spawn_new_pileup(self, start0, end0):
        return realign_rpileup.RealignerPileup(
            bam=self.bam, 
            chrom=self.chrom, start0=start0, end0=end0, 
            refver=self.refver, fasta=self.fasta, 
            aligner=self.aligner,
            init_df=True,
            verbose=self.verbose, logger=self.logger,
            **self.params,
        )

    def left_insert_new_pileup(self, width=1):
        new_pileup = self.spawn_new_pileup(
            start0=(self.start0 - width), end0=self.start0,
        )
        self.pileup_list.insert(0, new_pileup)

    def right_insert_new_pileup(self, width=1):
        new_pileup = self.spawn_new_pileup(
            start0=self.end0, end0=(self.end0 + width),
        )
        self.pileup_list.append(new_pileup)

    def extend_left(self, width):
        """Ignores MQ_limit, depth_limit, start0_limit, end0_limit, max_series_width attributes
        """
        assert width > 0

        target_start0 = self.start0 - width

        original_params = self.params.copy()
        self.params['max_series_width'] = np.inf
        self.params['start0_limit'] = target_start0
        self.params['MQ_limit'] = 0
        self.params['depth_limit'] = 0

        while True:
            self.left_insert_new_pileup(width=1)
            result_inactive, result_vcfspec, touched_width_limit = self.secure_left()
            if self.start0 <= target_start0:
                break

        self.params = original_params

#    def extend_left_old(self, width):
#        """Ignores MQ_limit, depth_limit, start0_limit, end0_limit,
#        max_series_width attribute"""
#        original_max_series_width = self.max_series_width
#        self.max_series_width = np.inf
#        original_start0_limit = self.start0_limit
#        self.start0_limit = -np.inf
#
#        target_start0 = self.start0 - width
#
#        rpileup_init_kwargs = self.rpileup_init_kwargs.copy()
#        rpileup_init_kwargs['start0_limit'] = target_start0
#        rpileup_init_kwargs['MQ_limit'] = 0
#        rpileup_init_kwargs['depth_limit'] = 0
#
#        while True:
#            new_pileup = RealignerPileup(
#                bam=self.bam, 
#                chrom=self.chrom, start0=(self.start0 - 1), end0=self.start0, 
#                refver=self.refver, fasta=self.fasta, 
#                verbose=self.verbose, logger=self.logger,
#                **rpileup_init_kwargs,
#            )
#            self.pileup_list.insert(0, new_pileup)
#            result_inactive, result_vcfspec, touched_width_limit = self.secure_left()
#            if self.start0 <= target_start0:
#                break
#
#        self.max_series_width = original_max_series_width
#        self.start0_limit = original_start0_limit

    def extend_right(self, width):
        """Ignores MQ_limit, depth_limit, start0_limit, end0_limit, max_series_width attributes
        """
        assert width > 0

        target_end0 = self.end0 + width

        original_params = self.params.copy()
        self.params['max_series_width'] = np.inf
        self.params['end0_limit'] = target_end0
        self.params['MQ_limit'] = 0
        self.params['depth_limit'] = 0

        while True:
            self.right_insert_new_pileup(width=1)
            result_inactive, result_vcfspec, touched_width_limit = self.secure_right()
            if self.end0 >= target_end0:
                break

        self.params = original_params

#    def extend_right_old(self, width):
#        """Ignores MQ_limit, depth_limit, start0_limit, end0_limit,
#        max_series_width attribute
#        """
#        original_max_series_width = self.max_series_width
#        self.max_series_width = np.inf
#        original_end0_limit = self.end0_limit
#        self.end0_limit = np.inf
#
#        target_end0 = self.end0 + width
#
#        rpileup_init_kwargs = self.rpileup_init_kwargs.copy()
#        rpileup_init_kwargs['end0_limit'] = target_end0
#        rpileup_init_kwargs['MQ_limit'] = 0
#        rpileup_init_kwargs['depth_limit'] = 0
#
#        while True:
#            new_pileup = RealignerPileup(
#                bam=self.bam, 
#                chrom=self.chrom, start0=self.end0, end0=(self.end0 + 1), 
#                refver=self.refver, fasta=self.fasta, 
#                verbose=self.verbose, logger=self.logger,
#                **rpileup_init_kwargs,
#            )
#            self.pileup_list.append(new_pileup)
#            result_inactive, result_vcfspec, touched_width_limit = self.secure_right()
#            if self.end0 >= target_end0:
#                break
#
#        self.max_series_width = original_max_series_width
#        self.end0_limit = original_end0_limit

    # init helpers #
    def secure_left(self):
        self.logger.debug(f'Beginning RealignerPileupSeries secure_left')

        left_pileup = self.pileup_list[0]
        left_pileup.params = self.params.copy()
        left_pileup.params['end0_limit'] = left_pileup.end0
        while True:
            result_inactive, result_vcfspec = left_pileup.secure_margins()

            self.logger.debug(self.pileup_list)
            self.logger.debug(f'self.start0: {self.start0}')
            self.logger.debug(f'result_inactive: {result_inactive}')
            self.logger.debug(f'result_vcfspec: {result_vcfspec}')

            if (
                (result_inactive.left_okay and result_vcfspec.left_okay)
                or (
                    result_vcfspec.left_low_depth or 
                    result_vcfspec.left_low_MQ or 
                    result_vcfspec.touched_left_limit
                )
                or (self.width > self.params['max_series_width'])
            ):
                break
            else:
                assert result_vcfspec.touched_width_limit
                split_points = left_pileup.get_split_points()
                if len(split_points) == 0:
                    self.left_insert_new_pileup(width=1)
                else:
                    split_pileups = left_pileup.split(split_points[:1])
                    del self.pileup_list[0]
                    self.pileup_list = split_pileups + self.pileup_list

                left_pileup = self.pileup_list[0]
                left_pileup.params['end0_limit'] = left_pileup.end0

        touched_width_limit = self.width >= self.params['max_series_width']

        self.logger.debug(f'Finished RealignerPileupSeries secure_left')

        return result_inactive, result_vcfspec, touched_width_limit

    def secure_right(self):
        self.logger.debug(f'Beginning RealignerPileupSeries secure_right')

        right_pileup = self.pileup_list[-1]
        right_pileup.params = self.params.copy()
        right_pileup.params['start0_limit'] = right_pileup.start0
        while True:
            result_inactive, result_vcfspec = right_pileup.secure_margins()

            self.logger.debug(self.pileup_list)
            self.logger.debug(f'self.end0: {self.end0}')
            self.logger.debug(f'result_inactive: {result_inactive}')
            self.logger.debug(f'result_vcfspec: {result_vcfspec}')

            if (
                (result_inactive.right_okay and result_vcfspec.right_okay)
                or (
                    result_vcfspec.right_low_depth or 
                    result_vcfspec.right_low_MQ or
                    result_vcfspec.touched_right_limit
                ) 
                or (self.width > self.params['max_series_width'])
            ):
                break
            else:
                assert result_vcfspec.touched_width_limit
                split_points = right_pileup.get_split_points()
                if len(split_points) == 0:
                    self.right_insert_new_pileup(width=1)
                else:
                    split_pileups = right_pileup.split(split_points[-1:])
                    del self.pileup_list[-1]
                    self.pileup_list.extend(split_pileups)

                right_pileup = self.pileup_list[-1]
                right_pileup.params['start0_limit'] = right_pileup.start0

        touched_width_limit = self.width >= self.params['max_series_width']

        self.logger.debug(f'Finished RealignerPileupSeries secure_right')

        return result_inactive, result_vcfspec, touched_width_limit

#    def secure_left_old(self):
#        self.logger.debug(f'Beginning RealignerPileupSeries secure_left')
#
#        left_pileup = self.pileup_list[0]
#        left_pileup.end0_limit = left_pileup.end0
#        while True:
#            result_inactive, result_vcfspec = left_pileup.secure_margins()
#
#            self.logger.debug(self.pileup_list)
#            self.logger.debug(f'self.start0: {self.start0}')
#            self.logger.debug(f'result_inactive: {result_inactive}')
#            self.logger.debug(f'result_vcfspec: {result_vcfspec}')
#
#            if (
#                (result_inactive.left_okay and result_vcfspec.left_okay)
#                or (
#                    result_vcfspec.left_low_depth or 
#                    result_vcfspec.left_low_MQ or 
#                    result_vcfspec.touched_left_limit
#                )
#                or (self.width > self.max_series_width)
#            ):
#                break
#            else:
#                assert result_vcfspec.touched_width_limit
#                split_points = left_pileup.get_split_points()
#                if len(split_points) == 0:
#                    rpileup_init_kwargs = self.rpileup_init_kwargs.copy()
#                    new_pileup = RealignerPileup(
#                        bam=self.bam, 
#                        chrom=self.chrom, start0=(left_pileup.start0 - 1), end0=left_pileup.start0, 
#                        refver=self.refver, fasta=self.fasta, 
#                        verbose=self.verbose, logger=self.logger,
#                        **rpileup_init_kwargs,
#                    )
#                    self.pileup_list.insert(0, new_pileup)
#                else:
#                    split_pileups = left_pileup.split(split_points[:1])
#                    del self.pileup_list[0]
#                    self.pileup_list = split_pileups + self.pileup_list
#
#                left_pileup = self.pileup_list[0]
#                left_pileup.end0_limit = left_pileup.end0
#
#        touched_width_limit = self.width >= self.max_series_width
#
#        self.logger.debug(f'Finished RealignerPileupSeries secure_left')
#
#        return result_inactive, result_vcfspec, touched_width_limit
#
#    def secure_right_old(self):
#        self.logger.debug(f'Beginning RealignerPileupSeries secure_right')
#
#        right_pileup = self.pileup_list[-1]
#        right_pileup.start0_limit = right_pileup.start0
#        while True:
#            result_inactive, result_vcfspec = right_pileup.secure_margins()
#
#            self.logger.debug(self.pileup_list)
#            self.logger.debug(f'self.end0: {self.end0}')
#            self.logger.debug(f'result_inactive: {result_inactive}')
#            self.logger.debug(f'result_vcfspec: {result_vcfspec}')
#
#            if (
#                (result_inactive.right_okay and result_vcfspec.right_okay)
#                or (
#                    result_vcfspec.right_low_depth or 
#                    result_vcfspec.right_low_MQ or
#                    result_vcfspec.touched_right_limit
#                ) or
#                or (self.width > self.max_series_width)
#            ):
#                break
#            else:
#                assert result_vcfspec.touched_width_limit
#                split_points = right_pileup.get_split_points()
#                if len(split_points) == 0:
#                    rpileup_init_kwargs = self.rpileup_init_kwargs.copy()
#                    new_pileup = RealignerPileup(
#                        bam=self.bam, 
#                        chrom=self.chrom, start0=right_pileup.end0, end0=(right_pileup.end0 + 1), 
#                        refver=self.refver, fasta=self.fasta, 
#                        verbose=self.verbose, logger=self.logger,
#                        **rpileup_init_kwargs,
#                    )
#                    self.pileup_list.append(new_pileup)
#                else:
#                    split_pileups = right_pileup.split(split_points[-1:])
#                    del self.pileup_list[-1]
#                    self.pileup_list.extend(split_pileups)
#
#                right_pileup = self.pileup_list[-1]
#                right_pileup.start0_limit = right_pileup.start0
#
#        touched_width_limit = self.width >= self.max_series_width
#
#        self.logger.debug(f'Finished RealignerPileupSeries secure_right')
#
#        return result_inactive, result_vcfspec, touched_width_limit

    def get_splittable_region_best(self):
        # without splitting contig vcfspec
        return pr.concat(
            [
                rpileup.get_vcfspec_margins_gr(
                    subseq_portion_threshold=(
                        rpileup.params['allele_portion_threshold'] * 2
                    ),
                    inverse=True,
                    split_contig_vcfspec=False,
                )
                for rpileup in self.pileup_list
            ]
        ).merge()

    def get_splittable_region_split_contig(self):
        return pr.concat(
            [
                rpileup.get_vcfspec_margins_gr(
                    subseq_portion_threshold=(
                        rpileup.params['allele_portion_threshold'] * 2
                    ),
                    inverse=True,
                    split_contig_vcfspec=True,
                )
                for rpileup in self.pileup_list
            ]
        ).merge()

    def get_splittable_region_inactive_runs(self):
        inactive_runs_gr_list = list()
        for rpileup in self.pileup_list:
            active_info_gr = rpileup.get_active_info_gr()
            inactive_gr = active_info_gr[~active_info_gr.Active]
            inactive_runs_gr_list.append(inactive_gr)
        result = pr.concat(inactive_runs_gr_list).merge()
        if result.empty:
            raise Exception(f'No inactive runs in this RealignerPileupSeries object.')
        return result

    def rearrange(self, start0_list, prepare_vcfspecs=True):
        """Mutates in-place"""
        def merger(left_pileup, right_pileup):
            left_pileup.merge(right_pileup, other_on_left=False)
            return left_pileup
        
        # sanity check
        if not (start0_list[0] >= self.start0 and start0_list[-1] <= self.end0):
            raise Exception(f'Input splitting range is out of PileupSeries object range.')
        
        # make new pileup_list
        split_ranges_gr = pr.from_dict(
            {
                'Chromosome': [self.chrom] * (len(start0_list) - 1),
                'Start': start0_list[:-1],
                'End': start0_list[1:],
                'Ranges_index': list(range(len(start0_list) - 1)),
            }
        )
        joined_gr = split_ranges_gr.join(self.get_gr())

        new_pileup_list = list()
        for ranges_index, subiter in itertools.groupby(
            (x[1] for x in joined_gr.df.iterrows()),
            key=(lambda x: x.Ranges_index),
        ):
            partial_pileups = list()
            for row in subiter:
                new_start0 = max(row.Start, row.Start_b)
                new_end0 = min(row.End, row.End_b)
                partial_pileups.append(
                    self.pileup_list[row.Self_index].subset(new_start0, new_end0)
                )
            new_pileup_list.append(functools.reduce(merger, partial_pileups))

        # prepare vcfspecs
        if prepare_vcfspecs:
            for pileup in new_pileup_list:
                pileup.prepare_vcfspecs()

        # result
        self.pileup_list = new_pileup_list

    @staticmethod
    def subseq_aln_to_cigartuples(alignment, left_is_empty, right_is_empty):
        # get active region cigartuples
        active_region_cigartuples, active_region_offset = alignhandler.alignment_to_cigartuples(
            alignment,
            match_as_78=False, del_as_skip=False, 
            left_ins_as_clip=left_is_empty, right_ins_as_clip=right_is_empty,
            remove_left_del=left_is_empty, remove_right_del=right_is_empty,
        )
        return active_region_cigartuples, active_region_offset

    def reduce_cigartuples_list_decorated(self, cigartuples_list_decorated):
        while True:
            cigartuples_list_decorated, reduced = self.softclip_reduce_cigartuples_list_decorated(cigartuples_list_decorated)
            if not reduced:
                break

        return functools.reduce(
            self.merge_two_cigartuples_plain,
            cigartuples_list_decorated,
        )[1]

    def softclip_reduce_cigartuples_list_decorated(self, cigartuples_list_decorated):
        reduced = False
        for idx, (left_item, right_item) in enumerate(tools.pairwise(cigartuples_list_decorated)):
            ref_range0_left, cigartuples_left, read_seq_left = left_item
            ref_range0_right, cigartuples_right, read_seq_right = right_item

            if (
                (ref_range0_right is not None)
                and (len(cigartuples_left) == 0)
                and (cigartuples_right[0][0] == 4)
                and (alignhandler.get_target_length(cigartuples_right) == len(ref_range0_right))
            ):  # Exceptional case where softclip faces empty side (221221)
                reduced = True
                merged_items = self.merge_two_cigartuples_right_clip_left_empty(
                    left_item, right_item
                )
                break
            elif (
                (ref_range0_left is not None)
                and (len(cigartuples_right) == 0)
                and (cigartuples_left[-1][0] == 4)
                and (alignhandler.get_target_length(cigartuples_left) == len(ref_range0_left))
            ):  # Exceptional case where softclip faces empty side (221221)
                reduced = True
                merged_items = self.merge_two_cigartuples_left_clip_right_empty(
                    left_item, right_item
                )
                break

        if reduced:
            result_cigartuples_list_decorated = (
                cigartuples_list_decorated[:idx]
                + [merged_items]
                + cigartuples_list_decorated[(idx + 2):]
            )
        else:
            result_cigartuples_list_decorated = cigartuples_list_decorated

        return result_cigartuples_list_decorated, reduced

    def merge_two_cigartuples_helper(self, read_seq_left, read_seq_right, ref_range0_left, ref_range0_right):
        new_read_seq = read_seq_left + read_seq_right
        if (ref_range0_left is None) or (ref_range0_right is None):
            new_ref_range0 = None
        else:
            new_ref_range0 = range(ref_range0_left.start, ref_range0_right.stop)

        return new_read_seq, new_ref_range0

    def merge_two_cigartuples_right_clip_left_empty(self, left_item, right_item):
        ref_range0_left, cigartuples_left, read_seq_left = left_item
        ref_range0_right, cigartuples_right, read_seq_right = right_item
        
        ref_seq = self.fasta.fetch(
            self.chrom, 
            ref_range0_right.start - int(cigartuples_right[0][1] + 20),
            ref_range0_right.start,
        )
        alignment = realign_base.main_aligner(
            target=ref_seq,
            query=read_seq_right[:cigartuples_right[0][1]],
            aligner=self.aligner,
            reverse_align=True,
        )
        new_cigartuples, offset = self.subseq_aln_to_cigartuples(
            alignment, left_is_empty=True, right_is_empty=False,
        )

        new_cigartuples = new_cigartuples + cigartuples_right[1:]

        new_read_seq, new_ref_range0 = self.merge_two_cigartuples_helper(
            read_seq_left, read_seq_right, ref_range0_left, ref_range0_right,
        )

        return new_ref_range0, new_cigartuples, new_read_seq

    def merge_two_cigartuples_left_clip_right_empty(self, left_item, right_item):
        ref_range0_left, cigartuples_left, read_seq_left = left_item
        ref_range0_right, cigartuples_right, read_seq_right = right_item
        
        ref_seq = self.fasta.fetch(
            self.chrom, 
            ref_range0_left.stop,
            ref_range0_left.stop + int(cigartuples_left[-1][1] + 20),
        )
        alignment = realign_base.main_aligner(
            target=ref_seq,
            query=read_seq_left[-cigartuples_left[-1][1]:],
            aligner=self.aligner,
            reverse_align=False,
        )

        new_cigartuples, offset = self.subseq_aln_to_cigartuples(
            alignment, left_is_empty=False, right_is_empty=True,
        )

        new_cigartuples = cigartuples_left[:-1] + new_cigartuples

        new_read_seq, new_ref_range0 = self.merge_two_cigartuples_helper(
            read_seq_left, read_seq_right, ref_range0_left, ref_range0_right,
        )

        return new_ref_range0, new_cigartuples, new_read_seq

    def merge_two_cigartuples_plain(self, left_item, right_item):
        ref_range0_left, cigartuples_left, read_seq_left = left_item
        ref_range0_right, cigartuples_right, read_seq_right = right_item
        
        if len(cigartuples_left) == 0 or len(cigartuples_right) == 0:
            new_cigartuples = cigartuples_left + cigartuples_right
            #return cigartuples_left + cigartuples_right
        else:
            if cigartuples_left[-1][0] == cigartuples_right[0][0]:
                new_cigartuples = (
                    cigartuples_left[:-1]
                    + [(cigartuples_left[-1][0], cigartuples_left[-1][1] + cigartuples_right[0][1])]
                    + cigartuples_right[1:]
                )
            else:
                new_cigartuples = cigartuples_left + cigartuples_right

        new_read_seq, new_ref_range0 = self.merge_two_cigartuples_helper(
            read_seq_left, read_seq_right, ref_range0_left, ref_range0_right,
        )

        return new_ref_range0, new_cigartuples, new_read_seq

    def save_subseq_alignments(self):
        for pileup in self.pileup_list:
            pileup.save_subseq_alignments()

    def set_realigned_reads(self):
        self.realigned_reads = dict()
        for read_uid, read in self.read_store.items():
            # select relevant pileups 
            relevant_pileups = list()
            for pileup in self.pileup_list:
                if read_uid in pileup.seq_df.index:
                    relevant_pileups.append(pileup)
            if len(relevant_pileups) == 0:
                # During 'subset' method of rpileup, out-of-range reads remain in 'read_store'
                continue

            # split cigartuples
            before_cigartuples, within_cigartuples, after_cigartuples = alignhandler.split_cigar(
                cigartuples=read.cigartuples, 
                reference_start0=read.reference_start,
                split_range0=range(self.start0, self.end0),
                trailing_queryonly_to_right=False,
            )

            before_empty = not bool(before_cigartuples)
            after_empty = not bool(after_cigartuples)

            # make cigartuples from alignment within each pileup
            offset_list = list()
            cigartuples_list = list()  # before, within1, within2, ..., after
            ref_range0_list = list()  # before, within1, within2, ..., after

            def helper(pileup, left_is_empty, right_is_empty):
                cigartuples, offset = self.subseq_aln_to_cigartuples(
                    alignment=pileup._subseq_alignments[read_uid], 
                    left_is_empty=left_is_empty,
                    right_is_empty=right_is_empty,
                )
                offset_list.append(offset)
                cigartuples_list.append(cigartuples)
                ref_range0_list.append(pileup.range0)

            cigartuples_list.append(before_cigartuples)
            ref_range0_list.append(None)  # ref_range0 is None for before_cigartuples

            if len(relevant_pileups) == 1:
                helper(relevant_pileups[0], before_empty, after_empty)
            else:
                helper(relevant_pileups[0], before_empty, False)
                for pileup in relevant_pileups[1:-1]:
                    helper(pileup, False, False)
                helper(relevant_pileups[-1], False, after_empty)

            cigartuples_list.append(after_cigartuples)
            ref_range0_list.append(None)

            # make partial read sequences
            partial_read_seqs = list()
            end_idx = 0
            for cigartuples in cigartuples_list:
                start_idx = end_idx
                end_idx += alignhandler.get_query_length(cigartuples)
                partial_read_seqs.append(read.query_sequence[start_idx:end_idx])

            cigartuples_list_decorated = list(
                zip(ref_range0_list, cigartuples_list, partial_read_seqs)
            )

            realigned_cigartuples = self.reduce_cigartuples_list_decorated(
                cigartuples_list_decorated
            )

            # get new read start position
            if not before_empty:
                new_reference_start = read.reference_start
            else:
                new_reference_start = relevant_pileups[0].start0 + offset_list[0]

            # make new read object
            realigned_read = read.__copy__()
            realigned_read.reference_start = new_reference_start
            realigned_read.cigartuples = realigned_cigartuples

            #####################################################

            readhandler.set_NMMD(realigned_read, fasta=self.fasta)
            self.realigned_cigar_sanity_check(read, realigned_read, realigned_cigartuples)
            self.realigned_reads[read_uid] = realigned_read

    def write_realigned_reads(self, bam_path, padding=0):
        # make header
        hdr = pysam.AlignmentHeader.from_references(
            reference_names=self.fasta.references,
            reference_lengths=self.fasta.lengths,
        )
        with pysam.AlignmentFile(bam_path, mode='wb', header=hdr) as in_bam:
            # write non-realigned reads
            for read in readhandler.get_fetch(
                self.bam,
                self.chrom,
                self.start0 - padding,
                self.end0 + padding,
            ):
                uid = readhandler.get_uid(read)
                if uid not in self.realigned_reads.keys():
                    in_bam.write(read)
            # write realigned reads
            for read in self.realigned_reads.values():
                in_bam.write(read)
        # sort and index
        bameditor.sort_and_index(bam_path)

    def get_result_vcfspecs(self, as_components=True, MQ_threshold=40): 
        return list(
            itertools.chain.from_iterable(
                rpileup.get_result_vcfspecs(as_components=as_components, MQ_threshold=MQ_threshold)
                for rpileup in self.pileup_list
            )
        )

    @staticmethod
    def realigned_cigar_sanity_check(original_read, realigned_read, realigned_cigartuples):
        if len(original_read.query_sequence) != alignhandler.get_query_length(realigned_cigartuples):
            raise realign_base.RealignmentCigarError(
                f'Read query sequence length and cigar length differs.\n'
                f'Original read:\n{original_read.to_string()}\n'
                f'Realigned read:\n{realigned_read.to_string()}\n'
            )



#!/usr/bin/env python3
 
#Author: Erik Bergstrom

#Contact: ebergstr@eng.ucsd.edu

from __future__ import print_function
import os
import sys
import time
import re
import argparse

start_time = time.time()

def save_tsb (chromosome_string_path, transcript_path, output_path):
	'''
	Creates binary files that contain the transcriptional information at a given 
	base. The transcriptional information entails Transcribed, Untranscribed, 
	Bi-directionally transcribed, and Non-transcribed. These files are required to
	simulate transcriptional strand bias.

	Input:
		chromosome_string_path -> path to chromosomes saved in individual text files
			   transcript_path -> path to transcript data. The following should be used as the format
								  (additional information may be saved, however it will not
								  be used when creating the files):

		Gene stable ID  Transcript stable ID    Chromosome/scaffold name    Strand  Transcript start (bp)   Transcript end (bp)

				   output_path -> path where the binary transcript files are saved.

	Output:
		Binary files that contain the transciptional information at each base for
		each chromosome. The files are saved as "[chrom]_192.txt"

	'''

	# Retrieves the transcript file 
	transcript_files = [f for f in os.listdir(transcript_path) if not f.startswith('.')]

	if len(transcript_files) < 3:
		# Instantiates required flag variables
		initial_flag = True
		initial_chrom = None
		for file in transcript_files:
			file_name = file.split("_")
			name = file_name[0]

			# Skips hidden files
			if name[0] == '.':
				pass
			else:

				# Sorts the transcript file by chromosome and transcript range
				#os.system("sort -t $'\t' -k 3,3n -k 3,3 -k 5,5 -k 6,6 " + transcript_path+file + " -o " + transcript_path+file)
				with open(transcript_path + file) as f:
					lines = [line.strip().split() for line in f]
				output = open(transcript_path + file, 'w')
				for line in sorted(lines, key = lambda x: (['I','II','III','IV','V','chrI','chrII','chrIII','chrIV','chrV','chrVI','chrVII','chrVIII','chrIX','chrX','chrXI','chrXII','chrXIII','chrXIV','chrXV','chrXVI','X','Y','1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24',
															'25','26','27','28','29','30','31','32','33','34','35','36','37','38','39', 'MT', 'M', 'MtDNA', 'chrM'].index(x[2]), int(x[4]), int(x[5]))):
					print('\t'.join(line), file=output)
				output.close()
				with open (transcript_path + file) as f:
					#next(f)

					# Saves all transcriptional information for a chromosome into a unique file
					for lines in f:
						line = lines.strip()
						line_split = line.split()
						chrom = line_split[2]

						if initial_flag:
							initial_chrom = chrom
							initial_flag = False
							out = open (transcript_path+chrom + "_transcripts.txt", 'w')

						if chrom == initial_chrom:
							print(line, file=out)

						else:
							out.close()
							initial_chrom = chrom
							out = open(transcript_path+chrom+"_transcripts.txt",'w')
					out.close()

			# Removes the initial transcript file
			# os.system("rm " + transcript_path + file)
			os.remove(transcript_path + file)

	# Retrieves the chromosome based list of transcript files 
	transcript_files = [f for f in os.listdir(transcript_path) if not f.startswith('.')]
	if '.DS_Store' in transcript_files:
		transcript_files.remove('.DS_Store')
	for files in transcript_files:
		with open(transcript_path + files) as f:
			lines = [line.strip().split() for line in f]

		output = open(transcript_path + files, 'w')

		for line in sorted(lines, key = lambda x: (int(x[4]))):
			print('\t'.join(line), file=output)

		output.close()

	print("Creating the transcriptional reference files now. This may take awhile...")

	first = True
	chrom_total = len(transcript_files)
	z = 1
	for files in transcript_files:
		file_name = files.split("_")
		chrom = file_name[0]
		if "chr" in chrom:
			chrom = chrom[3:]
		
		if chrom == '.DS':
			pass
		else:
			# try:
			if True:
				if os.path.exists(output_path + chrom + ".txt"):
					continue

				else:
					# Create a binary file for the output.
					outFile = open (output_path + chrom + ".txt", "wb")


					# Instantiates all of the required parameters.
					transcripts = None
					bi_start = 0
					bi_end= None
					I_chrom = None
					I_T_start = None
					I_T_end = None
					I_strand = None
					location = 0
					chrom_length = None

					# Save the length for the current chromosome.
					with open (chromosome_string_path + chrom + ".txt") as f:
						chrom_string = f.readline().strip()
						chrom_length = len(chrom_string)

					# Establishes a unique binary number for each transcript type.
					Non_transcribed_A = '00000000' # 0
					Non_transcribed_C = '00000001' # 1
					Non_transcribed_G = '00000010' # 2
					Non_transcribed_T = '00000011' # 3
					Transcribed_A = '00000100'     # 4
					Transcribed_C = '00000101'     # 5
					Transcribed_G = '00000110'     # 6
					Transcribed_T = '00000111'     # 7
					Untranscribed_A = '00001000'   # 8
					Untranscribed_C = '00001001'   # 9
					Untranscribed_G = '00001010'   # 10
					Untranscribed_T = '00001011'   # 11
					BiDirectional_A = '00001100'   # 12
					BiDirectional_C = '00001101'   # 13
					BiDirectional_G = '00001110'   # 14
					BiDirectional_T = '00001111'   # 15
					Non_transcribed_N = '00010000' # 16
					Transcribed_N = '00010001'     # 17
					Untranscribed_N = '00010010'   # 18 
					BiDirectional_N = '00010011'   # 19

					# Organizes the binary/integer formats for the data structures.
					nN_A = int(Non_transcribed_A, 2)
					nN_C = int(Non_transcribed_C, 2)
					nN_G = int(Non_transcribed_G, 2)
					nN_T = int(Non_transcribed_T, 2)
					nT_A = int(Transcribed_A, 2)
					nT_C = int(Transcribed_C, 2)
					nT_G = int(Transcribed_G, 2)
					nT_T = int(Transcribed_T, 2)
					nU_A = int(Untranscribed_A, 2)
					nU_C = int(Untranscribed_C, 2)
					nU_G = int(Untranscribed_G, 2)
					nU_T = int(Untranscribed_T, 2)
					nB_A = int(BiDirectional_A, 2)
					nB_C = int(BiDirectional_C, 2)
					nB_G = int(BiDirectional_G, 2)
					nB_T = int(BiDirectional_T, 2)
					
					nN_N = int(Non_transcribed_N, 2)
					nT_N = int(Transcribed_N, 2)
					nU_N = int(Untranscribed_N, 2)
					nB_N = int(BiDirectional_N, 2)

					byte_ref = {'N_A':bytes([nN_A]), 'N_C':bytes([nN_C]), 'N_G':bytes([nN_G]), 'N_T':bytes([nN_T]),
								'T_A':bytes([nT_A]), 'T_C':bytes([nT_C]), 'T_G':bytes([nT_G]), 'T_T':bytes([nT_T]),
								'U_A':bytes([nU_A]), 'U_C':bytes([nU_C]), 'U_G':bytes([nU_G]), 'U_T':bytes([nU_T]),
								'B_A':bytes([nB_A]), 'B_C':bytes([nB_C]), 'B_G':bytes([nB_G]), 'B_T':bytes([nB_T]),
								'N_N':bytes([nN_N]), 'T_N':bytes([nT_N]), 'U_N':bytes([nU_N]), 'B_N':bytes([nB_N])}

					stringTest = ''


					l = 0
					line_count = 1
					pointer = 0
					with open(transcript_path + files, 'r') as f:
						all_lines2 = f.readlines()
						all_lines = all_lines2[:]
			#            all_lines = all_lines2[1:]

						for lines in all_lines:
							
							# Handles the first line separately. 
							if pointer == 0:
								first_line = lines.split()
								location = int(first_line[4]) -1
								I_chrom = first_line[2][3:]
								I_T_start = int(first_line[4]) -1
								I_T_end = int(first_line[5]) - 1
								I_strand = first_line[3]
								
								# Saves Non-transcribed data up until
								# the first transcript.
								for i in range(0, location, 1):
									nuc = chrom_string[i].upper()
									outFile.write(byte_ref['N_' + nuc])
									l += 1
								pointer = 1

							# Handles all other lines after the first.
							else:
								first_line = lines.split()
								I_chrom = first_line[2][3:]
								I_T_start = int(first_line[4]) - 1
								I_T_end = int(first_line[5]) - 1
								I_strand = first_line[3]

								# Saves Non-transcribed data up until the 
								# next transcript.
								if I_T_start > location:
									for i in range(location, I_T_start, 1):
										nuc = chrom_string[i].upper()
										outFile.write(byte_ref['N_' + nuc])
										l += 1
									location = I_T_start
								
							# Reads the subsequent line to look for bi-directional
							# transcription sites.
							for line2 in all_lines[line_count:]:
								next_line = line2.split()
								c_chrom = next_line[2][3:]
								c_start = int(next_line[4]) - 1
								c_end= int(next_line[5]) - 1
								c_strand = next_line[3]
								
								# Breaks to the next line if the two transcripts 
								# don't overlap.
								if c_start > I_T_end:
									break

								# Checks if the transcripts are on opposite strands
								# if they overlap.    
								else:
									if c_strand != I_strand:
										bi_start = c_start
										if c_end > I_T_end:
											bi_end = I_T_end
										else:
											bi_end = c_end

									# Saves Un/Transcribed data up until the start of
									# the bi-directional transcription.
									if bi_start != 0 and bi_start > location:
										for i in range (location, bi_start, 1):
											if I_strand == '-1':
												nuc = chrom_string[i].upper()
												outFile.write(byte_ref['T_' + nuc])
												l += 1
											else:
												nuc = chrom_string[i].upper()
												outFile.write(byte_ref['U_' + nuc])
												l += 1
										location = bi_start

										# Saves Bi-directional data for the length
										# of the opposing transcripts overlap
										for i in range (location, bi_end, 1):
											nuc = chrom_string[i].upper()
											outFile.write(byte_ref['B_' + nuc])
											l += 1
										location = bi_end
									bi_start = 0
									bi_end = 0

							# Saves Un/Transcribed data up to the end of the current
							# transcript if data has not already been saved for these bases.
							if I_T_end > location:
								for i in range(location, I_T_end, 1):
									if I_strand == '-1':
										nuc = chrom_string[i].upper()
										outFile.write(byte_ref['T_' + nuc])
										l += 1
									else:
										nuc = chrom_string[i].upper()
										outFile.write(byte_ref['U_' + nuc])
										l +=1
								location = I_T_end
							line_count += 1

					# Save Non-transcribed data for the remainder of the chromosome after
					# the final transcript is analyzed         
					if location < chrom_length:
						for i in range(location, chrom_length,1):
							nuc = chrom_string[i].upper()
							outFile.write(byte_ref['N_' + nuc]) 
							l += 1        
						
					outFile.close()
					if first:
						print("The transcript reference file has been created for Chromosome: " +chrom + " (" + str(z) + "/" + str(chrom_total) + ")")
						first = False
					else:
						print("                                                               " +chrom + " (" + str(z) + "/" + str(chrom_total) + ")")
					z += 1
			#except:
			 #   print(chrom + " has been skipped. This chromosome is not supported.")

	end_time = time.time()
	print("Transcript files created.\n Job took: ", end_time-start_time, " seconds")

def main ():
	parser = argparse.ArgumentParser(description="Provide the necessary arguments to create the transcriptional strand bias reference strings.")
	parser.add_argument("--genome", "-g",help="Provide a reference genome. (ex: GRCh37, GRCh38, mm10)")
	args=parser.parse_args()
	genome = args.genome

	#script_dir = os.getcwd()
	#ref_dir = re.sub('\/scripts$', '', script_dir)
	ref_dir, tail = os.path.split(os.path.dirname(os.path.abspath(__file__)))

	chromosome_string_path = ref_dir + "/references/chromosomes/chrom_string/" + genome + "/"
	transcript_path = ref_dir + "/references/chromosomes/transcripts/" + genome + "/"
	output_path = ref_dir + "/references/chromosomes/tsb/" + genome + "/"
	if os.path.exists(output_path) == False:
		os.makedirs(output_path)

	save_tsb(chromosome_string_path, transcript_path, output_path)

if __name__ == '__main__':
	main()



import re
import os
import argparse
import textwrap

import pysam

import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup
import handygenome.vcfeditor.merge as libmerge
import handygenome.vcfeditor.indexing as indexing


def argument_parser(cmdargs):
    def sanity_check(args):
        e_msg = ('Invalid --isec-indices value. Please refer to the help '
                 'message.')
        if args.isec_indices is not None:
            if len(args.isec_indices) != len(args.infile_path_list):
                raise Exception(e_msg)
            elif not set(isec_indices).issubset({'0', '1'}):
                raise Exception(e_msg)

    def postprocess(args):
        if args.isec_indices is not None:
            args.isec_indices = [int(x) for x in args.isec_indices]

    parser_dict = workflow.init_parser()

    # required
    workflow.add_infilelist_arg(parser_dict['required'], required=True)
    workflow.add_outfile_arg(parser_dict['required'], required=True, 
                             must_not_exist='ask')
    workflow.add_fasta_arg(parser_dict['required'], required=True)
    parser_dict['required'].add_argument(
        '--mode', required=True, choices=('isec', 'union'),
        help=f'Must be "isec" (which means intersection) or "union".')

    # optional
    parser_dict['optional'].add_argument(
        '--isec-indices', dest='isec_indices', required=False,
        help=textwrap.dedent(f"""\
            (Only applied for intersection) A string composed of 0 or 1,
            with the same length as the number of input files. Files 
            marked with 0 are excluded and those with 1 are included. 
            If not set, all samples are included in intersection."""))

    # flag
    parser_dict['flag'].add_argument(
        '--remove-infoformat', dest='remove_infoformat', action='store_true',
        help=f'If set, all INFO and FORMAT data are removed from the output.')

    # others
    workflow.add_logging_args(parser_dict)
    workflow.add_index_arg(parser_dict)
    workflow.add_outfmt_arg(parser_dict['optional'], required=False)

    args = parser_dict['main'].parse_args(cmdargs)
    sanity_check(args)
    postprocess(args)

    return args


def main(cmdargs):
    args = argument_parser(cmdargs)
    logger = toolsetup.setup_logger(args)

    libmerge.main_file(
        infile_path_list=args.infile_path_list,
        outfile_path=args.outfile_path, fasta_path=args.fasta_path,
        remove_infoformat=args.remove_infoformat,
        isec=(args.mode == 'isec'), isec_indices=args.isec_indices,
        mode_pysam=args.mode_pysam, outfile_must_not_exist='no',
        logger=logger)

    if not args.donot_index:
        indexing.index_vcf(args.outfile_path)



import collections

import numpy as np
import pandas as pd
import scipy.stats
import scipy.signal
import scipy.linalg
import matplotlib as mpl
import matplotlib.pyplot as plt

#import handygenome.deco as deco
#import handygenome.logutils as logutils


class NoPeakError(Exception):
    pass


class DensityGenerationFailure(Exception):
    pass


class PeaksBase(collections.UserDict):
    @property
    def npeaks(self):
        return len(self['peak_xs'])

    def get_height_cutoff(self):
        if 'height' in self['find_peaks_kwargs'].keys():
            return self['find_peaks_kwargs']['height'][0]
        else:
            return None

    def subset(self, selector):
        """Values of 'peak_indexes' are not changed."""
        srcdict = dict()

        # first level value
        for key in (
            'peak_indexes',
            'peak_xs',
            'peak_ys',
            'ndata',
        ):
            if key in self:
                srcdict[key] = self[key][selector]

        # second level
        for key in (
            'peak_properties',
            'xcoord_values',
        ):
            if key in self:
                srcdict[key] = dict()
                for subkey, subval in self[key].items():
                    srcdict[key][subkey] = subval[selector]

        # do not subset
        for key in (
            'data',
            'find_peaks_kwargs',
            'xs',
            'ys',
            'density',
            'bin_edges',
            'bin_midpoints',
            'hist',
        ):
            if key in self:
                srcdict[key] = self[key]

        result = self.__class__(srcdict)
        return result

    #def get_highest_peak_index(self):
    #    max_peak_idx = np.argmax(self['peak_properties']['peak_heights'])
    #    return self['peak_indexes'][max_peak_idx]

    #############
    # filtering #
    #############

    def remove_contained_peaks(self):
        peak_properties = self['peak_properties']
        margins = [
            (x, y) for (x, y) in
            zip(
                peak_properties['left_ips'],
                peak_properties['right_ips'],
            )
        ]

        idxs_to_retain = list()
        for idx, query_margin in enumerate(margins):
            target_iterator = (
                x[1] for x in enumerate(margins) if x[0] != idx
            )
            contained = any(
                (
                    (query_margin[0] >= target_margin[0])
                    and (query_margin[1] <= target_margin[1])
                )
                for target_margin in target_iterator
            )
            if not contained:
                idxs_to_retain.append(idx)

        return self.subset(idxs_to_retain)

    def filter_ndata(self, n):
        self.subset(self['ndata'] >= n)

    def plot_common(self, ax, peakline_kwargs, draw_ndata=False):
        height_cutoff = self.get_height_cutoff()
        if height_cutoff is not None: 
            ax.axhline(height_cutoff, color='green', lw=2, alpha=0.4)

        for x, height in zip(
            self['peak_xs'], 
            self['peak_properties']['peak_heights'],
        ):
            ax.axvline(x, **peakline_kwargs)

        if draw_ndata:
            ylim = ax.get_ylim()
            y = ylim[1] + 0.05 * (ylim[1] - ylim[0])
            for (x, ndata) in zip(self['peak_xs'], self['ndata']):
                ax.text(x, y, str(ndata), va='center', ha='left', rotation=90)

    def postprocess(self, reltoheight=None):
        if reltoheight is not None:
            self.postprocess_width_reltoheight(reltoheight)
        self.postprocess_to_xcoords()
        self.postprocess_ndata()

    def postprocess_ndata(self):
        left_borders = self['xcoord_values']['left_ips']
        right_borders = self['xcoord_values']['right_ips']

        gt_left = self['data'][:, np.newaxis] >= left_borders
        lt_right = self['data'][:, np.newaxis] <= right_borders

        self['ndata'] = np.logical_and(gt_left, lt_right).sum(axis=0)

    def postprocess_to_xcoords(self):
        def convert_point(peak_properties_key):
            xs_indexes = self['peak_properties'][peak_properties_key]
            xs = self['xs']
            portions = xs_indexes / len(xs)
            return xs[0] + (xs[-1] - xs[0]) * portions

        def convert_length(peak_properties_key):
            lengths = self['peak_properties'][peak_properties_key]
            xs = self['xs']
            portions = lengths / len(xs)
            return (xs[-1] - xs[0]) * portions

        xcoord_values = dict()
        for key in (
            'left_bases', 
            'right_bases',

            'left_ips', 
            'right_ips',

            'left_ips_reltoheight', 
            'right_ips_reltoheight',
        ):
            if key in self['peak_properties']:
                xcoord_values[key] = convert_point(key)

        for key in (
            'widths',
            'widths_reltoheight',
        ):
            if key in self['peak_properties']:
                xcoord_values[key] = convert_length(key)

        self['xcoord_values'] = xcoord_values

    def postprocess_width_reltoheight(self, reltoheight):
        peak_properties = self['peak_properties']
        prominence_data = (
            peak_properties['prominences'],
            peak_properties['left_bases'],
            peak_properties['right_bases'],
        )
        rel_heights = (
            (peak_properties['peak_heights'] * reltoheight)
            / peak_properties['prominences']
        )
        width_results = list()
        for idx, (x_index, rel) in enumerate(
            zip(self['peak_indexes'], rel_heights)
        ):
            partial_result = scipy.signal.peak_widths(
                self['ys'], 
                [x_index], 
                rel_height=rel,
                prominence_data=(
                    np.atleast_1d(peak_properties['prominences'][idx]),
                    np.atleast_1d(peak_properties['left_bases'][idx]),
                    np.atleast_1d(peak_properties['right_bases'][idx]),
                ),
            )
            width_results.append(partial_result)

        peak_properties['widths_reltoheight'] = np.concatenate(
            [x[0] for x in width_results]
        )
        peak_properties['width_heights_reltoheight'] = np.concatenate(
            [x[1] for x in width_results]
        )
        peak_properties['left_ips_reltoheight'] = np.concatenate(
            [x[2] for x in width_results]
        )
        peak_properties['right_ips_reltoheight'] = np.concatenate(
            [x[3] for x in width_results]
        )

    def undecal(self, x):
        '''peak_indexes are not modified'''
        def undecal_value(arr):
            return (2 * x) - arr

        def modify_pos(arr, gt_selector):
            arr[gt_selector] = undecal_value(arr[gt_selector])

        def modify_borders(lt_arr, rt_arr, gt_selector):
            undecaled_rt_values = undecal_value(rt_arr[gt_selector])
            undecaled_lt_values = undecal_value(lt_arr[gt_selector])
            lt_arr[gt_selector] = undecaled_rt_values
            rt_arr[gt_selector] = undecaled_lt_values

        gt_selector = self['peak_xs'] > x

        # peak_xs
        modify_pos(self['peak_xs'], gt_selector)

        # peak_properties, xcoord_values
        peak_properties = self['peak_properties']
        xcoord_values = self['xcoord_values']
        for lt_key, rt_key in (
            ('left_bases', 'right_bases'),
            ('left_ips', 'right_ips'),
            ('left_ips_reltoheight', 'right_ips_reltoheight'),
        ):
            if lt_key in peak_properties:
                modify_borders(
                    peak_properties[lt_key], 
                    peak_properties[rt_key], 
                    gt_selector,
                )
                modify_borders(
                    xcoord_values[lt_key], 
                    xcoord_values[rt_key], 
                    gt_selector,
                )


class HistPeaks(PeaksBase):
    def plot(
        self, 
        figsize=None, 
        ax=None, 
        peakline_kwargs=dict(),
        bar_kwargs=dict(),
    ):
        peakline_kwargs = (
            {'lw': 0.5, 'color': 'blue'}
            | peakline_kwargs
        )
        bar_kwargs = (
            {'fill': 'tab:blue'}
            | bar_kwargs
        )

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = None

        x = self['bin_edges'][:-1]
        height = self['hist']
        width = np.diff(self['bin_edges'])
        ax.bar(x, height, width, align='edge', **bar_kwargs)

        self.plot_common(ax, peakline_kwargs)

        return fig, ax


class DensityPeaks(PeaksBase):
    def plot(
        self, 
        figsize=None, 
        ax=None, 
        peakline_kwargs=dict(),
        plot_kwargs=dict(),
    ):
        peakline_kwargs = (
            {'lw': 0.5, 'color': 'red'}
            | peakline_kwargs
        )
        plot_kwargs = (
            {'color': 'tab:red'}
            | plot_kwargs
        )

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = None

        ax.plot(self['xs'], self['ys'], **plot_kwargs)

        self.plot_common(ax, peakline_kwargs)

        return fig, ax


def find_hist_peaks(
    data, 

    weights=None, 
    limit=None, 
    reltoheight=None,
    inverse=False,

    bins=10, 
    **find_peaks_kwargs,
):
    data, weights, find_peaks_kwargs = findpeaks_arghandler(
        data, weights, limit, find_peaks_kwargs,
    )

    hist, bin_edges = np.histogram(
        data, bins=bins, weights=weights, density=True,
    )
    bin_midpoints = 0.5 * (bin_edges[:-1] + bin_edges[1:])

    if inverse:
        findpeaks_input = np.max(hist) - hist
    else:
        findpeaks_input = hist
    peaks, peak_properties = scipy.signal.find_peaks(
        findpeaks_input, **find_peaks_kwargs
    )

    if len(peaks) == 0:
        raise NoPeakError('Peaks could not be found from histogram')
        #return None
    else:
        peakresult = HistPeaks(
            {
                'data': data,
                'find_peaks_kwargs': find_peaks_kwargs,
                'peak_indexes': peaks,
                'peak_xs': bin_midpoints[peaks],
                'peak_ys': peak_properties['peak_heights'],
                'peak_properties': peak_properties,
                'bin_edges': bin_edges,
                'bin_midpoints': bin_midpoints,
                'hist': hist,
                'xs': bin_midpoints,
                'ys': hist,
            }
        )
        peakresult.postprocess(reltoheight=reltoheight)
        return peakresult


def find_density_peaks(
    data, 

    weights=None, 
    limit=None, 
    reltoheight=None,
    inverse=False,

    xs=100, 
    bw_method=None, 
    **find_peaks_kwargs,
):
    data, weights, find_peaks_kwargs = findpeaks_arghandler(
        data, weights, limit, find_peaks_kwargs,
    )

    try:
        density = scipy.stats.gaussian_kde(
            data, weights=weights, bw_method=bw_method,
        )
    except Exception as exc:
        msg = f'data: {data}; weights: {weights}'
        raise DensityGenerationFailure(msg) from exc

#    except scipy.linalg.LinAlgError as exc:
#        if str(exc) == 'singular matrix':
#            logutils.log(
#                f'Density generation failed.', 
#                add_locstring=True, 
#                level='warn',
#            )
#            return None
#        else:
#            raise

    if isinstance(xs, int):
        xs = np.linspace(data.min(), data.max(), xs)
        #xs = np.linspace(np.quantile(data, 0.01), np.quantile(data, 0.99), xs)
    ys = density(xs)

    if inverse:
        findpeaks_input = np.max(ys) - ys
    else:
        findpeaks_input = ys
    peaks, peak_properties = scipy.signal.find_peaks(
        findpeaks_input, **find_peaks_kwargs
    )

    if len(peaks) == 0:
        raise NoPeakError('Peaks could not be found from density')
        #return None
    else:
        peakresult = DensityPeaks(
            {
                'data': data,
                'find_peaks_kwargs': find_peaks_kwargs,
                'peak_indexes': peaks,
                'peak_xs': xs[peaks],
                'peak_ys': peak_properties['peak_heights'],
                'peak_properties': peak_properties,
                'density': density,
                'xs': xs,
                'ys': ys,
            }
        )
        peakresult.postprocess(reltoheight=reltoheight)
        return peakresult


def findpeaks_arghandler(data, weights, limit, find_peaks_kwargs):
    data = np.asarray(data)
    if data.ndim != 1:
        raise Exception(f'Input data must be 1-d.')

    if weights is None:
        weights = np.repeat(1, len(data))
    else:
        weights = np.asarray(weights)

    if limit is not None:
        selector = np.logical_and(data > limit[0], data < limit[1])
        data = data[selector]
        weights = weights[selector]

    # find_peaks_kwargs
    find_peaks_kwargs = (
        {
            'height': (None, None), 
            'width': (None, None),
            'rel_height': 0.5,
        }
        | find_peaks_kwargs
    )

    return data, weights, find_peaks_kwargs


def draw_peaks(
    data=None, 

    omit_density=False,

    histpeaks=None,
    densitypeaks=None,

    hist_kwargs=dict(),
    density_kwargs=dict(), 
):
    # set kwargs
    hist_kwargs = (
        dict(bins=30)
        | hist_kwargs
    )
    density_kwargs = (
        dict()
        | density_kwargs
    )

    # main
    if histpeaks is None:
        histpeaks = find_hist_peaks(data, **hist_kwargs)
    if (densitypeaks is None) and (not omit_density):
        try:
            densitypeaks = find_density_peaks(data, **density_kwargs)
        except DensityGenerationFailure:
            omit_density = True

    fig, ax = plt.subplots()
    histpeaks.plot(
        ax=ax, 
    )
    if not omit_density:
        densitypeaks.plot(
            ax=ax, 
        )

    return fig, ax
